import pprint
from dataclasses import asdict, dataclass, field
from typing import List, Optional

from ML_management.singleton_pattern import Singleton


@dataclass(init=True)
class DebugJobResult:
    job_name: str
    job_id: int
    artifacts: Optional[str] = None
    metrics: Optional[str] = None
    params: Optional[str] = None
    models: List[dict] = field(default_factory=list)
    buckets: List[dict] = field(default_factory=list)

    def __repr__(self):
        printer = pprint.PrettyPrinter(width=80, indent=4)
        return "\nResult of the job: \n" + printer.pformat(asdict(self))


@dataclass(init=True, repr=True)
class DebugJobLogContext(DebugJobResult, metaclass=Singleton):
    job_name: Optional[str] = None
    job_id: Optional[int] = None
    old_python_path: list = None
    old_sys_modules: list = None
    extra_sys_path: List[str] = field(default_factory=list)

    def clear(self):
        self.__init__()

    def get_result(self) -> DebugJobResult:
        return DebugJobResult(
            job_id=self.job_id,
            job_name=self.job_name,
            artifacts=self.artifacts,
            metrics=self.metrics,
            params=self.params,
            models=self.models,
            buckets=self.buckets,
        )



"""Local collector."""
from ML_management.collectors.collector_pattern import CollectorPattern


class LocalCollector(CollectorPattern):
    """Local collectors."""

    def set_data(self, local_path: str) -> str:
        """Set suitable data."""
        return local_path

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        schema = {
            "type": "object",
            "properties": {"local_path": {"type": "string"}},
            "required": ["local_path"],
            "additionalProperties": False,
        }

        return schema



"""Dummy collectors."""
from ML_management.collectors.collector_pattern import CollectorPattern


class DummyCollector(CollectorPattern):
    """Dummy collectors."""

    def set_data(self, *_args, **_kwargs):
        """Set suitable data."""
        return ""

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        schema = {"type": "object", "properties": {}, "required": [], "additionalProperties": False}

        return schema



from ML_management import variables


def set_server_url(url: str) -> None:
    """
    Set server URL.

    If you set the URL using this function,
    it takes precedence over the URL from the environment variable 'server_url'.

    Parameters
    ==========
    url: str
        Server url.
    Returns
    =======
    None
    """
    variables.server_url = url


def set_mlm_credentials(login: str, password: str) -> None:
    """
    Set login and password for mlmanagement.

    If you set the credentials using this function,
    it takes precedence over the credentials from the environment variables.

    Parameters
    ==========
    login: str
        Your mlm login.
    passwrod: str
        Yor mlm password.
    Returns
    =======
    None
    """
    variables.mlm_login = login
    variables.mlm_password = password


def get_server_url() -> str:
    """Get the current server URL."""
    return variables.get_server_url()


def set_no_cache_load(no_cache: bool):
    """
    Set flag no cache for download artifacts.

    If you set no_cache using this function,
    it takes precedence over the environment variable NO_CACHE.

    Parameters
    ==========
    no_cache: bool

    Returns
    =======
    None
    """
    variables.NO_CACHE = no_cache


def set_local_registry_path(registry_path: str):
    """
    Set local registry path.

    If you set local registry path using this function,
    it takes precedence over the environment variable LOCAL_REGISTRY_PATH.

    Parameters
    ==========
    registry_path: str
        registry path.
    Returns
    =======
    None
    """
    variables.LOCAL_REGISTRY_PATH = registry_path


def set_debug_registry_path(registry_path: str):
    """
    Set debug registry path.

    Parameters
    ==========
    registry_path: str
        registry path.
    Returns
    =======
    None
    """
    variables.DEBUG_REGISTRY_PATH = registry_path


def get_debug() -> bool:
    """Get the current debug flag."""
    return variables.DEBUG


def set_debug(debug: bool):
    """
    Set debug flag.

    If you set the flag using this function,
    it takes precedence over the credentials from the environment variables.

    Parameters
    ==========
    debug: bool
        Your mlm login.
    Returns
    =======
    None
    """
    variables.DEBUG = debug



"""Collector for fetching texts marked with ceritain topic, as well as negative examples."""
import os
from typing import List

import pandas as pd
from sgqlc.endpoint.http import HTTPEndpoint
from sgqlc.operation import Operation

from ML_management.collectors.collector_pattern import CollectorPattern
from ML_management.collectors.topic_markers.api_schema import Query


class TopicMarkersCollector(CollectorPattern):
    """Collector class for fetching texts marked with ceritain topic, as well as negative examples."""

    def __init__(self):
        # can not import GRAPHQL_URL from server's env_variables file
        self.graphql_url = os.environ.get("GRAPHQL_URL", "http://localhost:9000/graphql")
        self.endpoint = HTTPEndpoint(self.graphql_url)
        # TODO for now, recreate csv file every time at a hardcoded location

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        schema = {
            "type": "object",
            "properties": {
                "local_path": {"type": "string"},
                "marker": {"type": "string"},
                "start_date": {"type": "integer"},
                "end_date": {"type": "integer"},
            },
            "required": [
                "marker",
            ],
            "additionalProperties": False,
        }

        return schema

    def __save_texts(self, texts: List[str], labels: List[int], path: str):
        """Save texts and labels to csv by passed path."""
        # path is supposed to be a full path to file
        data_dict = {
            "text": texts,
            "label": labels,
        }
        dataframe = pd.DataFrame(data_dict)
        # shuffle everything
        dataframe = dataframe.sample(frac=1).reset_index(drop=True)
        # this should work correctly with commas and quotes in text...
        dataframe.to_csv(path, index=False)

    def __gql_query(self, form: dict) -> dict:
        """Make GraphQL query to fetch new data from IE module."""
        op = Operation(Query)

        query = op.list_text_from_document_with_marker(form=form)
        query.marker_text()
        query.not_marker_text()

        data = self.endpoint(op)
        return data

    def set_data(self, *_args, **kwargs):
        """Fetch texts with passed marker, serialize into given dir and return path."""
        local_path = kwargs.get("local_path", "topic_markers.csv")
        form = {
            "marker": kwargs["marker"],
        }
        if "start_date" in kwargs:
            form["startDate"] = kwargs["start_date"]
        if "end_date" in kwargs:
            form["endDate"] = kwargs["end_date"]

        raw_data = self.__gql_query(form)
        marker_texts = raw_data["data"]["listTextFromDocumentWithMarker"]["markerText"]
        not_marker_texts = raw_data["data"]["listTextFromDocumentWithMarker"]["notMarkerText"]
        # TODO do we raise if there are no examples whatsoever?
        all_labels = [1] * len(marker_texts) + [0] * len(not_marker_texts)
        all_texts = marker_texts + not_marker_texts
        self.__save_texts(texts=all_texts, labels=all_labels, path=local_path)

        return local_path



"""Collectors."""
from typing import Callable, Dict, Generic, ItemsView, Type, TypeVar

from ML_management.collectors.collector_pattern import CollectorPattern

_Value = TypeVar("_Value")


class ModelTypeFactory(Dict[str, _Value], Generic[_Value]):
    """Class for wrap call."""

    def __init__(self, items: Dict[str, Callable[[], _Value]]) -> None:
        """Init."""
        super().__init__(items)

    def __getitem__(self, k: str) -> _Value:
        """Get item."""
        return super().__getitem__(k)()

    def items(self) -> ItemsView[str, Callable[[], _Value]]:
        """Items."""
        return super().items()


def _get_s3() -> Type[CollectorPattern]:
    """Get s3."""
    from ML_management.collectors.s3.s3collector import S3Collector

    return S3Collector


def _get_topic_marker() -> Type[CollectorPattern]:
    """Get topic_marker."""
    from ML_management.collectors.topic_markers.topic_markers_collector import TopicMarkersCollector

    return TopicMarkersCollector


def _get_dummy() -> Type[CollectorPattern]:
    """Get dummy."""
    from ML_management.collectors.dummy.dummy_collector import DummyCollector

    return DummyCollector


def _get_local() -> Type[CollectorPattern]:
    """Get local."""
    from ML_management.collectors.local.local_collector import LocalCollector

    return LocalCollector


DATA_COLLECTORS = ModelTypeFactory(
    {
        "s3": _get_s3,
        "topic_marker": _get_topic_marker,
        "dummy": _get_dummy,
        "local": _get_local,
    }
)

DATA_REMOTE_COLLECTORS = ModelTypeFactory(
    {
        "s3": _get_s3,
        "topic_marker": _get_topic_marker,
        "dummy": _get_dummy,
    }
)



"""Dataset loader template for custom dataset loader."""
from abc import ABC, abstractmethod


class DatasetLoaderPattern(ABC):
    """Define dataset loader.

    Attributes
    ----------
    artifacts : str
        Local path to artifacts.
        That parameters will be set automatically in job before the 'get_dataset' func would be executed.
    data_path: str
        A path to data to be loaded.

    """

    def __init__(self):
        """Init dataset loader class."""
        self.artifacts: str
        self.data_path = None

    @abstractmethod
    def get_dataset(self, **dataset_params):
        """
        Return dataset.

        To get data_path use self.data_path parameter, which also will be set in the job.
        'dataset_params' are dataset_loader parameters. One has to define it as ordinary kwargs
        with type annotation.
        """
        raise NotImplementedError



"""S3 Manager for operations with s3 data."""
import concurrent.futures
import datetime
import math
import mimetypes
import os
import posixpath
import random
import shutil
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from http import HTTPStatus
from pathlib import Path
from typing import List, Literal, Optional, Union

import boto3
import magic
from boto3.s3.transfer import TransferConfig
from botocore.awsrequest import AWSRequest, AWSResponse
from botocore.config import Config
from botocore.exceptions import ClientError
from pydantic import BaseModel
from tqdm.autonotebook import tqdm

from ML_management import variables
from ML_management.local_debug.local_logger import LocalLogger
from ML_management.mlmanagement import get_debug
from ML_management.mlmanagement.load_api import _untar_folder
from ML_management.mlmanagement.log_api import _tar_folder
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.s3.utils import (
    FileLock,
    get_bucket_info,
    get_bucket_size,
    get_upload_paths,
    get_upload_size,
)
from ML_management.session import AuthSession
from ML_management.variables import LOCAL_REGISTRY_PATH

KB = 1024
MB = KB * KB
MAX_PARTS = 10000
MAX_KEYS = 1000
DEFAULT_MULTIPART_CHUNKSIZE = 8 * MB
S3_DATA = "s3_data"


class S3BucketNotFoundError(Exception):
    """Define Bucket Not Found Exception."""

    def __init__(self, bucket: str):
        self.bucket = bucket
        self.message = f'Bucket "{bucket}" does not exist'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3BucketNotFoundError, (self.bucket,))


class S3ObjectNotFoundError(Exception):
    """Define Version Not Found Exception."""

    def __init__(self, path: str, bucket: str):
        self.path = path
        self.bucket = bucket
        self.message = f'Object "{path}" is not found in "{bucket}" bucket'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3ObjectNotFoundError, (self.path, self.bucket))


class S3UpdateVisibilityError(Exception):
    """Define Version Not Found Exception."""

    def __init__(self, bucket: str, code: int, message: str):
        self.code = code
        self.bucket = bucket
        self.message = f"Error status {code} when trying to update bucket {bucket} visibility with message '{message}'."
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3UpdateVisibilityError, (self.bucket, self.code, self.message))


class FileIsNotTarError(Exception):
    """Define exception when tar is expected but not given."""

    def __init__(self, filename: str):
        self.filename = filename
        self.message = f"Expected tar file, but file {self.filename} does not have '.tar' extension"
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (FileIsNotTarError, (self.filename,))


class AmbiguousFileChoiceError(Exception):
    """Define exception when there is not exactly one file in bucket to choice."""

    def __init__(self, number_of_files: int):
        self.number_of_files = number_of_files
        if not number_of_files:
            appendix = " bucket is empty"
        else:
            appendix = "there is more than one file in bucket. Specify one file name with 'remote_paths' parameter"
        self.message = f"Expected one tar file in bucket, but {appendix}."
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (AmbiguousFileChoiceError, (self.number_of_files,))


class TimestampInterval(BaseModel):
    """Input wrapper for timestamp interval."""

    start: Optional[int]
    end: Optional[int]

    def __init__(self, start: Optional[int] = None, end: Optional[int] = None):
        if start and end and end < start:
            raise ValueError(
                f"The specified time interval from {start} to {end} is invalid: at least "
                f"one value should be present, and the end parameter should not exceed start."
            )
        super().__init__(start=start, end=end)

    def to_params(self):
        params = {}

        if self.start:
            params["creation_datetime_start_interval"] = self.start
        if self.end:
            params["creation_datetime_end_interval"] = self.end

        return params


class BucketFilterSettings(BaseModel):
    """Filter settings for list buckets."""

    name: Optional[str] = None
    visibility: Optional[str] = None
    creation_datetime_interval: Optional[TimestampInterval] = None
    owner_ids: Optional[list[str]] = None

    def to_params(self):
        params = {}
        for key, value in self.dict().items():
            if key == "creation_datetime_interval" and value:
                params.update(self.creation_datetime_interval.to_params())
            if value:
                params[key] = value
        return params


class ObjectInfo(BaseModel):
    """
    Info about the object from the bucket.

    Attributes:
        name (str): Name of the object.
        last_modified (datetime): Creation date of the object.
        size (int): Size in bytes of the object
    """

    name: str
    last_modified: datetime.datetime
    size: int


class BucketInfo(BaseModel):
    """
    Info about the bucket.

    Attributes:
        name (str): Name of the bucket.
        prefix (str): the prefix passed in the request.
        max_keys (int): The maximum number of keys returned in the response.
        key_count (int): The number of keys returned with this request.
        next_continuation_token (Optional[str]): If сontinuation_token was sent with the request,
            it is included in the response. You can use the returned next_continuation_token
            for pagination of the list response.
        contents (List[ObjectInfo]): Info about each object returned.

    """

    name: str
    prefix: str
    max_keys: int
    key_count: int
    next_continuation_token: Optional[str]
    contents: List[ObjectInfo]


class S3Manager:
    """Manager for operations with s3 objects."""

    def __init__(self) -> None:
        """Init creds."""
        self.default_url = variables.get_s3_gateway_url()
        self.default_access_key_id, self.default_secret_access_key = variables.get_s3_credentials()
        self.session = AuthSession()

    def list_objects(
        self, bucket: str, max_keys: int = MAX_KEYS, prefix: str = "", continuation_token: Optional[str] = None
    ) -> BucketInfo:
        """
        Returns some(up to 1,000) or all of the objects in a bucket.

        Parameters
        ----------
        bucket: str
            Name of bucket you want to show.
        prefix: str = ""
            the file or folder will be show to the bucket with this prefix.
            Example: if prefix = "foo", file "bar" would be show to the bucket as "foo/bar"
        max_keys: int
            The maximum number of keys returned in the response. If 0 is passed, it will return all objects.
            Defaults to 1000.

        continuation_token:Optional[str] = None
            If ContinuationToken was sent with the request, it is included in the response.
            You can use the returned ContinuationToken for pagination of the list response.

        Returns
        -------
        BucketInfo instance with meta information.
        """
        with self._get_sync_boto_client() as s3_client:
            if max_keys > 0:
                kwargs = {"Bucket": bucket, "Prefix": prefix, "MaxKeys": max_keys}
                if continuation_token is not None:
                    kwargs["ContinuationToken"] = continuation_token
                response = s3_client.list_objects_v2(**kwargs)
                contents = [
                    ObjectInfo(name=content["Key"], last_modified=content["LastModified"], size=content["Size"])
                    for content in response.get("Contents", [])
                ]
                key_count = response["KeyCount"]
                max_keys = response["MaxKeys"]
                next_continuation_token = response.get("NextContinuationToken", None)
            else:
                next_continuation_token = None
                paginator = s3_client.get_paginator("list_objects_v2")
                page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)
                contents = []
                key_count = 0
                for page in page_iterator:
                    key_count += page["KeyCount"]
                    contents.extend(
                        ObjectInfo(name=content["Key"], last_modified=content["LastModified"], size=content["Size"])
                        for content in page["Contents"]
                    )
        return BucketInfo(
            name=bucket,
            prefix=prefix,
            max_keys=max_keys,
            key_count=key_count,
            contents=contents,
            next_continuation_token=next_continuation_token,
        )

    @contextmanager
    def _get_sync_boto_client(self):
        s3_client = boto3.client(
            service_name="s3",
            use_ssl=True,
            endpoint_url=posixpath.join(self.default_url, "s3/"),
            aws_access_key_id=self.default_access_key_id,
            aws_secret_access_key=self.default_secret_access_key,
            config=Config(request_checksum_calculation="when_required"),
        )
        event_system = s3_client.meta.events
        event_system.register("before-sign.s3.*", self._add_auth_cookies)
        event_system.register("after-call.s3.*", self._update_auth_cookies)
        yield s3_client

    def list_buckets(self, filter_settings: Optional[BucketFilterSettings] = None) -> List[str]:
        """Get list names of available buckets."""
        params = None
        if filter_settings:
            params = filter_settings.to_params()
        with self.session.get(posixpath.join(self.default_url, "list-buckets"), params=params) as response:
            buckets_info: list[dict[str, str]] = response.json()["buckets"]
            return [bucket["name"] for bucket in buckets_info]

    def create_bucket(
        self,
        name: str,
        visibility: Union[Literal["private", "public"], VisibilityOptions],
    ) -> None:
        """Create new bucket with specified name and visibility."""
        visibility = VisibilityOptions(visibility)
        with self.session.post(
            posixpath.join(self.default_url, "create-bucket"), json={"name": name, "visibility": visibility.value}
        ) as response:
            if response.status_code != HTTPStatus.CREATED:
                raise RuntimeError(f"Failed to create bucket: {response.text}")

    def update_bucket_visibility(
        self,
        bucket: str,
        new_visibility: Union[Literal["private", "public"], VisibilityOptions],
    ) -> None:
        """Update bucket visibility."""
        new_visibility = VisibilityOptions(new_visibility)
        if bucket not in self.list_buckets():
            raise S3BucketNotFoundError(bucket)
        with self.session.post(
            posixpath.join(self.default_url, "update-bucket-auth", bucket),
            json={"visibility": new_visibility.value},
        ) as response:
            if not response.is_success:
                raise S3UpdateVisibilityError(bucket, response.status_code, response.text)

    def delete_bucket(self, bucket: str, verbose: bool = True, max_workers: Optional[int] = None) -> None:
        """Delete bucket."""
        self.delete_by_prefix(bucket=bucket, prefix="", verbose=verbose, max_workers=max_workers)
        with self._get_sync_boto_client() as s3_client:
            s3_client.delete_bucket(Bucket=bucket)

    def delete_by_prefix(self, bucket: str, prefix: str, verbose: bool = True, max_workers: Optional[int] = None):
        """Delete all objects with specified prefix in key.

        Parameters
        ----------
        bucket: str
            name of bucket you want to delete directory from.
        prefix: str
            path to your objects or dir
        max_workers: Optional[int] = None
            Maximum number of worker threads for parallel requests per file. Default: None
        verbose: bool = True
            If the option is set to True and upload_as_tar set to False,
            a progress bar with the number of uploaded files will be displayed.

            Example::

            |    Directories structure:
            |    ├───dockerfile
            |    ├───a.py
            |    ├───dir1
            |            ├───empty.txt
            |            └───dir2
            |                    ├───new.txt
            |                    ├───x.png
            |                    └───ipsum
            |                            └───y.png
            |    To delete dir1    use prefix="dir1"
            |    To delete dir2    use prefix="dir1/dir2"
            |    To delete new.txt use prefix="dir1/dir2/new.txt"
        """
        if bucket not in self.list_buckets():
            raise S3BucketNotFoundError(bucket)

        with self._get_sync_boto_client() as s3_client:
            paginator = s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

            total_size = 0
            for page in page_iterator:
                total_size += page["KeyCount"]
            page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

            with tqdm(
                total=total_size,
                disable=not verbose,
                unit_scale=True,
                unit_divisor=1024,
                unit="File",
            ) as pbar:
                futures = []
                with ThreadPoolExecutor(max_workers=max_workers) as thread_executor:
                    for page in page_iterator:
                        for obj in page.get("Contents", []):
                            object_key = obj.get("Key")
                            futures.append(
                                thread_executor.submit(
                                    self._delete_one_object,
                                    bucket=bucket,
                                    key=object_key,
                                    s3_client=s3_client,
                                    pbar=pbar,
                                )
                            )

                    tasks = concurrent.futures.wait(futures, return_when="FIRST_EXCEPTION")
                    try:
                        for error_task in tasks.done:
                            error_task.result()
                    except Exception as e:
                        for task in tasks.not_done:
                            task.cancel()
                        raise e

    def upload(
        self,
        local_path: str,
        bucket: str,
        upload_as_tar: bool = False,
        new_bucket_visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
        max_workers: Optional[int] = None,
        verbose: bool = True,
        prefix: str = "",
    ):
        """
        Upload directory to bucket.

        Parameters
        ----------
        local_path: str
            path to directory with files you want to upload.
        bucket: str
            name of bucket you want to upload to.
        upload_as_tar: bool = False
            If the option is set to True, the files will be uploaded as a single tar archive. Default: False
        max_workers: Optional[int] = None
            Maximum number of worker threads for parallel requests per file. Default: None
        verbose: bool = True
            If the option is set to True and upload_as_tar set to False,
            a progress bar with the number of uploaded files will be displayed.
        prefix: str = ""
            the file or folder will be uploaded to the bucket with this prefix.
            Example: if prefix = "foo", file "bar" would be uploaded to the bucket as "foo/bar"
        new_bucket_visibility: Union[Literal['private', 'public'], VisibilityOptions]
            Visibility of this bucket to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
            Defaults to PRIVATE.
        """
        new_bucket_visibility = VisibilityOptions(new_bucket_visibility)

        if get_debug():
            LocalLogger().upload_bucket(local_path, bucket, prefix)
            return
        prefix = prefix.strip("/")
        if upload_as_tar:
            self._upload_as_tar(
                local_path=local_path,
                bucket=bucket,
                prefix=prefix,
                new_bucket_visibility=new_bucket_visibility,
                verbose=verbose,
            )
        else:
            self._upload_files(
                local_path=local_path,
                bucket=bucket,
                prefix=prefix,
                new_bucket_visibility=new_bucket_visibility,
                verbose=verbose,
                max_workers=max_workers,
            )

    def set_data(
        self,
        *,
        local_path: Optional[str] = None,
        bucket: str,
        untar_data: bool = False,
        remote_paths: Optional[List[str]] = None,
        max_workers: Optional[int] = None,
        verbose: bool = True,
        sync: bool = True,
        clear_local: bool = False,
    ) -> str:
        """
        Set data.

        :type local_path: Optional[str]
        :param local_path: Local path to save data to.  Defaults to MLM_REGISTRY_PATH or S3_DATA when caching disable.

        :type bucket: string
        :param bucket: Bucket containing requested files.

        :type untar_data: bool
        :param untar_data: If the option is set to True,
            the tar file from the bucket will be downloaded and untarred into the folder.

        :type remote_paths: list(string)
        :param remote_paths: List of paths relative to passed bucket.  Each path
            can represent either a single file, or a folder.  If a path represents
            a folder (should end with a slash), then all contents of a folder are recursively downloaded.

        :type max_workers: Optional[int]
        :param max_workers: Maximum number of worker threads for parallel requests per file.

        :type verbose: bool
        :param verbose: Whether to disable the entire progressbar wrapper.

        :type sync: bool
        :param sync: Whether to synchronize local folder and remote s3 data or just download files again.

        :type clear_local: bool
        :param clear_local: Whether to clear local folder before downloading.
        """
        if local_path is None and not variables.NO_CACHE:
            local_path = os.path.join(LOCAL_REGISTRY_PATH, "row_data")
        elif local_path is None:
            local_path = S3_DATA

        if bucket not in self.list_buckets():
            raise S3BucketNotFoundError(bucket)

        print("Start to downloading bucket...")
        with FileLock(local_path, bucket):
            local_path = os.path.join(local_path, bucket)

            if os.path.exists(local_path) and clear_local:
                shutil.rmtree(local_path, ignore_errors=True)

            if untar_data:
                return self._download_data_tar(
                    local_path=local_path, bucket=bucket, remote_paths=remote_paths, verbose=verbose, sync=sync
                )

            return self._download_files(
                local_path=local_path,
                bucket=bucket,
                remote_paths=remote_paths,
                verbose=verbose,
                max_workers=max_workers,
                sync=sync,
            )

    def _delete_one_object(self, bucket: str, key: str, s3_client, pbar, retries: int = 3):
        # do retries if there are some network troubles, cause bucket and dir deleting depends on it,
        # if object was deleted earlier, s3_client.delete_object has no effect
        while retries > 0:
            try:
                s3_client.delete_object(Bucket=bucket, Key=key)
                pbar.update()
                break
            except Exception:
                retries -= 1
                if retries == 0:
                    raise
                time.sleep(random.randint(1, 9) * 1e-4)

    def _upload_as_tar(
        self, local_path: str, bucket: str, prefix: str, new_bucket_visibility: VisibilityOptions, verbose: bool
    ) -> None:
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(f"Path: {local_path} does not exist")

        total_size = get_upload_size(local_path)

        multipart_chunksize = math.ceil(total_size / MAX_PARTS)  # in bytes
        if multipart_chunksize < DEFAULT_MULTIPART_CHUNKSIZE:
            multipart_chunksize = DEFAULT_MULTIPART_CHUNKSIZE

        r, w = os.pipe()

        try:
            thread = threading.Thread(target=_tar_folder, args=(w, local_path), daemon=True)
            thread.start()
        except Exception as err:
            os.close(r)
            os.close(w)
            raise err
        with self._get_sync_boto_client() as s3_client:
            buckets = self.list_buckets()
            if bucket not in buckets:
                self.create_bucket(name=bucket, visibility=new_bucket_visibility)
            try:
                with tqdm(
                    total=total_size,
                    disable=not verbose,
                    unit_scale=True,
                    unit_divisor=1024,
                    unit="B",
                ) as pbar:
                    with open(r, "rb") as fileobj:
                        s3_client.upload_fileobj(
                            Fileobj=fileobj,
                            Bucket=bucket,
                            Key=os.path.join(prefix, f"{os.path.basename(local_path)}.tar"),
                            Callback=pbar.update,
                            ExtraArgs={"ContentType": "application/x-tar"},
                            Config=TransferConfig(multipart_chunksize=multipart_chunksize),
                        )
            finally:
                thread.join()

    def _upload_files(
        self,
        local_path: str,
        bucket: str,
        prefix: str,
        new_bucket_visibility: Union[Literal["private", "public"], VisibilityOptions],
        max_workers: Optional[int],
        verbose: bool = True,
    ):
        new_bucket_visibility = VisibilityOptions(new_bucket_visibility)
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(f"Path: {local_path} does not exist")
        with self._get_sync_boto_client() as s3_client:
            buckets = self.list_buckets()
            if bucket not in buckets:
                self.create_bucket(name=bucket, visibility=new_bucket_visibility)

            upload_paths, total_size = get_upload_paths(local_path)

            with tqdm(
                total=total_size,
                disable=not verbose,
                unit_scale=True,
                unit_divisor=1024,
                unit="B",
            ) as pbar:
                futures = []
                with ThreadPoolExecutor(max_workers=max_workers) as thread_executor:
                    for path in upload_paths:
                        media_type = mimetypes.guess_type(path.local_path)[0]
                        if not media_type:
                            media_type = magic.from_file(path.local_path, mime=True)

                        multipart_chunksize = math.ceil(path.size / MAX_PARTS)  # in bytes
                        if multipart_chunksize < DEFAULT_MULTIPART_CHUNKSIZE:
                            multipart_chunksize = DEFAULT_MULTIPART_CHUNKSIZE
                        futures.append(
                            thread_executor.submit(
                                s3_client.upload_file,
                                Filename=path.local_path,
                                Bucket=bucket,
                                Key=os.path.join(prefix, path.storage_path),
                                Callback=pbar.update,
                                ExtraArgs={"ContentType": media_type},
                                Config=TransferConfig(multipart_chunksize=multipart_chunksize),
                            )
                        )
                    tasks = concurrent.futures.wait(futures, return_when="FIRST_EXCEPTION")
                    try:
                        for error_task in tasks.done:
                            error_task.result()
                    except Exception as e:
                        for task in tasks.not_done:
                            task.cancel()
                        raise e

    def _set_data_sync(
        self,
        local_path: str,
        bucket: str,
        remote_paths: List[str],
        max_workers: Optional[int] = None,
        verbose: bool = True,
    ):
        """Synchronize bucket and local folder."""
        with self._get_sync_boto_client() as s3_client:
            paginator = s3_client.get_paginator("list_objects_v2")

            total_bucket_size, bucket_info = get_bucket_info(bucket, remote_paths, paginator)
            # get only local_paths with prefix in remote_paths
            local_files = []

            for remote_path in remote_paths:
                full_remote_path = os.path.join(local_path, remote_path)
                if Path(full_remote_path).is_file():
                    local_files.append(Path(full_remote_path))
                else:
                    local_files.extend(list(Path(full_remote_path).rglob("*")))  # breadth-first search

            with tqdm(
                total=total_bucket_size,
                unit_scale=True,
                unit_divisor=1024,
                disable=not verbose,
                unit="B",
            ) as pbar:
                futures = []
                with ThreadPoolExecutor(max_workers=max_workers) as thread_executor:
                    for path in reversed(local_files):
                        rel_path = path.relative_to(local_path)
                        if str(rel_path) not in bucket_info:
                            if path.is_file():
                                os.remove(path)
                            else:
                                # rm dir if empty
                                if not os.listdir(path):
                                    os.rmdir(path)
                            continue

                        local_last_modified = datetime.datetime.fromtimestamp(
                            os.path.getmtime(path), tz=datetime.timezone.utc
                        )
                        bucket_file_last_modified = bucket_info[str(rel_path)]["LastModified"].replace(
                            tzinfo=datetime.timezone.utc
                        )
                        if local_last_modified > bucket_file_last_modified:
                            pbar.update(bucket_info[str(rel_path)]["Size"])
                            bucket_info.pop(str(rel_path), None)
                            continue

                        bucket_info.pop(str(rel_path), None)

                        futures.append(
                            thread_executor.submit(
                                self._download_one_object,
                                bucket,
                                str(rel_path),
                                os.path.dirname(path),
                                path,
                                s3_client,
                                pbar.update,
                            )
                        )

                    for key in bucket_info.keys():
                        futures.append(
                            thread_executor.submit(
                                self._download_one_object,
                                bucket,
                                key,
                                os.path.dirname(os.path.join(local_path, key)),
                                os.path.join(local_path, key),
                                s3_client,
                                pbar.update,
                            )
                        )

                    tasks = concurrent.futures.wait(futures, return_when="FIRST_EXCEPTION")
                    try:
                        for error_task in tasks.done:
                            error_task.result()
                    except Exception as e:
                        for task in tasks.not_done:
                            task.cancel()
                        raise e

        return local_path

    def _download_data_tar(
        self,
        *,
        local_path: str,
        bucket: str,
        verbose: bool,
        remote_paths: Optional[List[str]] = None,
        sync: bool,
    ) -> str:
        remote_paths = remote_paths if remote_paths else None
        if remote_paths is not None and len(remote_paths) != 1:
            raise RuntimeError(f"Expected one tar object, but {len(remote_paths)} were given.")

        with self._get_sync_boto_client() as s3_client:
            if remote_paths is not None:
                tar_name = remote_paths[0]
                if not tar_name.endswith(".tar"):
                    raise FileIsNotTarError(tar_name)
            else:
                try:
                    # check that there is only one file in bucket
                    # If there is more than 1 file, raise exception about ambiguous choice
                    paginator = s3_client.get_paginator("list_objects_v2")
                    page_iterator = paginator.paginate(Bucket=bucket, Prefix="", PaginationConfig={"MaxItems": 2})
                    first_page = next(iter(page_iterator))
                    files = first_page.get("Contents", [])
                    if len(files) != 1:
                        raise AmbiguousFileChoiceError(number_of_files=len(files))
                    tar_name: str = files[0]["Key"]
                    tar_size: int = files[0]["Size"]
                    if not tar_name.endswith(".tar"):
                        raise FileIsNotTarError(tar_name)
                except ClientError as err:
                    if err.response["Error"]["Code"] == "NoSuchBucket":
                        raise S3BucketNotFoundError(bucket=bucket) from None
                    else:
                        raise

            try:
                head_object = s3_client.head_object(Bucket=bucket, Key=tar_name)
                tar_size = head_object["ContentLength"]

            except ClientError as err:
                if err.response["Error"]["Code"] == "NoSuchBucket":
                    raise S3BucketNotFoundError(bucket=bucket) from None
                if err.response["Error"]["Code"] == "404":
                    raise S3ObjectNotFoundError(path=tar_name, bucket=bucket) from None
                else:
                    raise

            local_path = os.path.join(local_path, tar_name.removesuffix(".tar"))

            if sync and os.path.exists(local_path):
                print("Bucket exist in local cache, synchronize with remote.")
                local_last_modified = min(
                    datetime.datetime.fromtimestamp(os.path.getatime(root), tz=datetime.timezone.utc)
                    for root, _, _ in os.walk(local_path)
                )
                if local_last_modified > head_object["LastModified"]:
                    with tqdm(
                        total=tar_size,
                        unit_scale=True,
                        unit_divisor=1024,
                        disable=not verbose,
                        unit="B",
                    ) as pbar:
                        pbar.update(tar_size)
                    return local_path
            if os.path.exists(local_path):
                shutil.rmtree(local_path, ignore_errors=True)

            return self._download_tar(
                local_path=local_path,
                s3_client=s3_client,
                tar_name=tar_name,
                tar_size=tar_size,
                bucket=bucket,
                verbose=verbose,
            )

    def _download_tar(
        self, local_path: str, tar_name: str, tar_size: int, bucket: str, verbose: bool, s3_client
    ) -> str:
        r, w = os.pipe()
        with open(r, "rb") as buff:
            try:
                thread = threading.Thread(target=_untar_folder, args=(buff, os.path.dirname(local_path)), daemon=True)
                thread.start()
            except Exception as err:
                os.close(r)
                os.close(w)
                raise err
            try:
                with open(w, "wb") as fileobj:
                    with tqdm(
                        total=tar_size,
                        unit_scale=True,
                        unit_divisor=1024,
                        disable=not verbose,
                        unit="B",
                    ) as pbar:
                        s3_client.download_fileobj(Fileobj=fileobj, Bucket=bucket, Key=tar_name, Callback=pbar.update)
            except ClientError as err:
                if err.response["Error"]["Code"] == "404":
                    raise S3ObjectNotFoundError(path=tar_name, bucket=bucket) from None
                raise
            finally:
                thread.join()
            return local_path

    def _download_files(
        self,
        *,
        local_path: str,
        bucket: str,
        remote_paths: Optional[List[str]] = None,
        max_workers: Optional[int],
        verbose: bool = True,
        sync: bool,
    ) -> str:
        remote_paths = remote_paths if remote_paths else [""]

        if sync and os.path.exists(local_path):
            print("Bucket exist in local cache, synchronize with remote.")
            return self._set_data_sync(local_path, bucket, remote_paths, max_workers, verbose)

        with self._get_sync_boto_client() as s3_client:
            paginator = s3_client.get_paginator("list_objects_v2")

            total_bucket_size = get_bucket_size(bucket, remote_paths, paginator)

            with tqdm(
                total=total_bucket_size,
                unit_scale=True,
                unit_divisor=1024,
                disable=not verbose,
                unit="B",
            ) as pbar:
                futures = []
                with ThreadPoolExecutor(max_workers=max_workers) as thread_executor:
                    for remote_path in remote_paths:
                        page_iterator = paginator.paginate(Bucket=bucket, Prefix=remote_path)
                        for page in page_iterator:
                            if page["KeyCount"] == 0 and remote_path != "":
                                raise S3ObjectNotFoundError(path=remote_path, bucket=bucket)
                            for obj in page.get("Contents", []):
                                file_path = obj.get("Key")

                                local_dir_path = os.path.join(local_path, posixpath.dirname(file_path))
                                local_file_path = os.path.join(local_path, file_path)
                                futures.append(
                                    thread_executor.submit(
                                        self._download_one_object,
                                        bucket,
                                        file_path,
                                        local_dir_path,
                                        local_file_path,
                                        s3_client,
                                        pbar.update,
                                    )
                                )

                    tasks = concurrent.futures.wait(futures, return_when="FIRST_EXCEPTION")
                    try:
                        for error_task in tasks.done:
                            error_task.result()
                    except Exception as e:
                        for task in tasks.not_done:
                            task.cancel()
                        raise e
            return local_path

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _add_auth_cookies(self, request: AWSRequest, **kwargs) -> None:  # noqa
        request.headers.add_header("Cookie", self.session._get_cookie_header())

    # arguments to callback are passed like kwargs, so kwargs must be present in signature
    def _update_auth_cookies(self, http_response: AWSResponse, **kwargs) -> None:  # noqa
        cookie_header = http_response.headers.get("set-cookie")
        if cookie_header is None:
            return
        cookies: list[str] = cookie_header.split("; ")
        for cookie in cookies:
            if "kc-access" not in cookie:
                continue
            _, new_access_token = cookie.split("=", maxsplit=1)
            self.session.cookies["kc-access"] = new_access_token
            break

    def _download_one_object(
        self, bucket: str, key: str, local_dir_path: str, local_file_path: str, s3_client, callback
    ):
        if not os.path.exists(local_dir_path):
            os.makedirs(local_dir_path, exist_ok=True)
        s3_client.download_file(bucket, key, local_file_path, Callback=callback)



"""Functions for internal usage."""
import hashlib
import json
import os
import re
from typing import Dict, List, Optional

from pydantic import BaseModel, field_validator

from ML_management.variables import FILENAME_FOR_INFERENCE_CONFIG

INIT_FUNCTION_NAME = "get_object"

HASH_IGNORE_PATH = ["__pycache__"]

HASH_CHUNK_SIZE = 4096 * 1024

model_name_pattern = re.compile("(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+")

valid_data_types = [
    "bool",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "int32",
    "int64",
    "float16",
    "float32",
    "float64",
]


class InferenceParams(BaseModel):
    name: str
    data_type: str
    dims: List[int]

    @field_validator("data_type")
    @classmethod
    def data_type_check(cls, value):
        assert value in valid_data_types, (
            f"{FILENAME_FOR_INFERENCE_CONFIG}: Every object in 'input' or 'output' list "
            f"should have one of the following data types: {valid_data_types}"
        )


class PredictConfig(BaseModel):
    cfg: Dict[str, List[InferenceParams]]

    @field_validator("cfg")
    @classmethod
    def cfg_check(cls, value):
        assert (
            "input" in value and "output" in value
        ), f"File {FILENAME_FOR_INFERENCE_CONFIG} should contain both 'input' and 'output' keys"


def is_model_name_valid(name: str):
    return model_name_pattern.fullmatch(name) is not None


def validate_predict_config(path: Optional[str]):
    if path is None or not os.path.isfile(path):
        raise RuntimeError(
            f"File {FILENAME_FOR_INFERENCE_CONFIG} was not found in artifacts. "
            "It is required, when create_venv_pack=True"
        )

    with open(path) as f:
        try:
            data = json.load(f)
        except Exception as err:
            raise RuntimeError(f"File {FILENAME_FOR_INFERENCE_CONFIG} should be valid json.") from err

    PredictConfig(cfg=data)


def hash_file(filepath: str) -> str:
    hash_sha256 = hashlib.sha256()
    with open(filepath, "rb") as f:
        for chunk in iter(lambda: f.read(HASH_CHUNK_SIZE), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


def calculate_hash_directory(directory: str) -> str:
    hash_sha256 = hashlib.sha256()

    for root, _, files in sorted(os.walk(directory)):
        if any(root.endswith(ignore) for ignore in HASH_IGNORE_PATH):
            continue
        for name in sorted(files):
            filepath = os.path.join(root, name)
            file_hash = hash_file(filepath)
            hash_sha256.update(os.path.relpath(filepath, directory).encode("utf-8"))
            hash_sha256.update(file_hash.encode("utf-8"))

    return hash_sha256.hexdigest()


def calculate_size(local_path: str) -> int:
    if os.path.isfile(local_path):
        return os.path.getsize(local_path)
    total_size = 0
    for root, _, files in os.walk(local_path):
        for name in files:
            filepath = os.path.join(root, name)
            total_size += os.path.getsize(filepath)

    return total_size



import json
import logging
import os
import shutil
import sys
import warnings
from typing import Optional, Union

from ML_management import variables
from ML_management.executor import BaseExecutor
from ML_management.graphql.schema import ExecutionJob
from ML_management.job.job import Job
from ML_management.local_debug.debug_job_result import DebugJobLogContext, DebugJobResult
from ML_management.local_debug.local_logger import LocalLogger
from ML_management.mlmanagement.backend_api import get_debug, set_debug
from ML_management.mlmanagement.load_api import _load_from_src, load_dataset, load_executor, load_model
from ML_management.model.model_type_to_methods_map import ModelMethodName
from ML_management.sdk.parameters import (
    AnyDatasetLoaderForm,
    AnyModelForm,
    DatasetLoaderForm,
    ModelForm,
    UploadAnyNewModelsForm,
    UploadOneNewModelForm,
)
from ML_management.variables import DEBUG_REGISTRY_PATH, LOCAL_REGISTRY_PATH

LOCAL_DEBUG_DIR = "mlm_registry"
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


class DebugContext:
    def __init__(self, job_name: str):
        self.job_name = job_name
        self.old_debug = True

    def __enter__(self):
        self.old_debug = get_debug()
        set_debug(True)
        DebugJobLogContext().job_name = self.job_name
        DebugJobLogContext().old_python_path = sys.path.copy()
        DebugJobLogContext().old_sys_modules = sys.modules.copy()
        DebugJobLogContext().job_id = LocalLogger().get_job_id()
        variables.secret_uuid = DebugJobLogContext().job_id
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_debug(self.old_debug)
        variables.secret_uuid = None
        sys.path = DebugJobLogContext().old_python_path
        for module in set(sys.modules) - set(DebugJobLogContext().old_sys_modules):
            if hasattr(sys.modules[module], "__file__"):
                if sys.modules[module].__file__ and any(
                    extra in sys.modules[module].__file__ for extra in DebugJobLogContext().extra_sys_path
                ):
                    sys.modules.pop(module)
        DebugJobLogContext().clear()


def add_ml_job_local(
    job_executor_name: str,
    executor_params: dict,
    models_pattern: Union[ModelForm, AnyModelForm],
    data_pattern: Union[DatasetLoaderForm, AnyDatasetLoaderForm],
    job_name: str,
    job_executor_version: Optional[int] = None,
    upload_models_params: Union[UploadOneNewModelForm, UploadAnyNewModelsForm, None] = None,
    gpu: bool = False,
    model_paths: Optional[Union[dict, str]] = None,
    dataset_loader_paths: Optional[Union[dict, str]] = None,
    executor_path: Optional[str] = None,
) -> DebugJobResult:
    """
    Create local job for debugging purposes.

    Parameters
    ----------
    job_executor_name: str
        Name of the executor that will execute the job.
    executor_params: Dict[str, ...]
        Dictionary of executor parameters.
        Example::

            {
                'executor_param1': 'value1',
                'executor_param2': 'value2',
                'executor_param3': 'value3',
                ...
            }

    models_pattern: Union[ModelForm, AnyModelForm]
        Necessary information for using the models.
    data_pattern: Union[DatasetLoaderForm, AnyDatasetLoaderForm]
        Necessary information for using the datasets.
    gpu: bool = False
        Whether to use GPU for this job or not. Default: False
    job_executor_version: Optional[int] = None
        Version of the executor that will execute the job. Default: None, "latest" version is used.
    upload_models_params: Union[UploadOneNewModelForm, UploadAnyNewModelsForm],
        Parameters to log new models. If you want to log new model or model version,
        you have to specify this parameter. Default: None.
        The value None means that no new model will be uploaded.
    job_name: Optional[str] = None
        Name of the created job.
    model_paths: Optional[Union[dict, str]] = None
        Local paths to the models.
    dataset_loader_paths: Optional[Union[dict, str]] = None
        Local paths to the dataset loaders.
    executor_path: Optional[Union[dict, str]] = None
        Local path to the executor.

    Returns
    -------
    DebugJobResult
        Result of the Job.
    """
    if not get_debug():
        warnings.warn("The debug environment variable must be set to True.")

    if not isinstance(models_pattern, (ModelForm, AnyModelForm)):
        raise TypeError("Parameter models must have type ModelForm or AnyModelForm.")
    if not isinstance(data_pattern, (DatasetLoaderForm, AnyDatasetLoaderForm)):
        raise TypeError("Parameter data_pattern must have type DatasetLoaderForm or AnyDatasetLoaderForm.")
    if (
        upload_models_params is not None
        and not isinstance(upload_models_params, UploadOneNewModelForm)
        and not isinstance(upload_models_params, UploadAnyNewModelsForm)
    ):
        raise TypeError(
            "Parameter upload_params must have type UploadOneNewModelForm or UploadAnyNewModelsForm or be None."
        )
    if upload_models_params is None:
        upload_models_params = UploadAnyNewModelsForm(upload_models_params=[])
    logger.info("Start Job")
    with DebugContext(job_name):
        models = models_pattern.serialize_gql()
        data_params = data_pattern.serialize_gql()
        upload_params = upload_models_params.serialize_gql()

        if executor_path:
            preload_executor = _load_from_src(executor_path, None)

        else:
            preload_executor = load_executor(job_executor_name, job_executor_version, install_requirements=True)
            job_executor_version = preload_executor.metainfo.version
            preload_executor = preload_executor.loaded_object

        if job_executor_version is None:
            job_executor_version = 0
        preload_models = {}
        if model_paths and not isinstance(model_paths, dict):
            model_paths = {BaseExecutor.DEFAULT_ROLE: model_paths}
        elif not model_paths:
            model_paths = {}

        model_roles_expected = {model["role"] for model in models}
        if set(model_paths.keys()) - model_roles_expected:
            raise ValueError(f"Extra model roles passed: {set(model_paths.keys()) - model_roles_expected}")

        preload_dataset_loaders = {}
        if dataset_loader_paths and not isinstance(dataset_loader_paths, dict):
            dataset_loader_paths = {BaseExecutor.DEFAULT_ROLE: dataset_loader_paths}
        elif not dataset_loader_paths:
            dataset_loader_paths = {}

        dataset_loader_roles_expected = {dataset_loader["role"] for dataset_loader in data_params}
        if set(dataset_loader_paths.keys()) - dataset_loader_roles_expected:
            raise ValueError(
                f"Extra dataset loader roles passed: {set(model_paths.keys()) - dataset_loader_roles_expected}"
            )
        models_roles = set()
        for model in models:
            role = model["role"]
            models_roles.add(role)
            kwargs_for_init = _get_init_parameters_from_dict(model["modelParams"]["listModelMethodParams"])

            if role in upload_params:
                model["uploadParams"] = upload_params[role]
                if model["uploadParams"]["newModelName"] is None:
                    model["uploadParams"]["newModelName"] = model["modelParams"]["modelVersionChoice"]["name"]
            else:
                model["uploadParams"] = None
            if role in model_paths:
                preload_models[role] = _load_from_src(model_paths[role], kwargs_for_init)
                if not model["modelParams"]["modelVersionChoice"]["version"]:
                    model["modelParams"]["modelVersionChoice"]["version"] = 0
                continue
            preload_model = load_model(
                model["modelParams"]["modelVersionChoice"]["name"],
                model["modelParams"]["modelVersionChoice"]["version"],
                install_requirements=True,
                kwargs_for_init=kwargs_for_init,
            )
            model["modelParams"]["modelVersionChoice"]["version"] = preload_model.metainfo.version
            preload_models[role] = preload_model.loaded_object
        if not set(upload_params.keys()).issubset(models_roles):
            raise ValueError("Roles in upload params must be subset of roles in model params.")

        for dataset_loader in data_params:
            role = dataset_loader["role"]
            kwargs_for_init = _get_init_parameters_from_dict(
                dataset_loader["dataParams"]["listDatasetLoaderMethodParams"]
            )
            if role in dataset_loader_paths:
                preload_dataset_loaders[role] = _load_from_src(dataset_loader_paths[role], kwargs_for_init)

                if not dataset_loader["dataParams"]["datasetLoaderVersionChoice"]["version"]:
                    dataset_loader["dataParams"]["datasetLoaderVersionChoice"]["version"] = 0
                continue
            preload_dataset_loader = load_dataset(
                dataset_loader["dataParams"]["datasetLoaderVersionChoice"]["name"],
                dataset_loader["dataParams"]["datasetLoaderVersionChoice"]["version"],
                install_requirements=True,
                kwargs_for_init=kwargs_for_init,
            )
            dataset_loader["dataParams"]["datasetLoaderVersionChoice"][
                "version"
            ] = preload_dataset_loader.metainfo.version
            preload_dataset_loaders[role] = preload_dataset_loader.loaded_object

        ep = {
            "executorMethodParams": {"methodName": "execute", "methodParams": json.dumps(executor_params)},
            "executorVersionChoice": {"name": job_executor_name, "version": job_executor_version},
        }

        params = {
            "listRoleModelParams": models,
            "listRoleDataParams": data_params,
            "executorParams": ep,
            "resources": {"cpus": 1, "memoryPerNode": 1000, "gpuNumber": 1 if gpu else 0, "gpuType": None},
        }
        job = ExecutionJob(
            {"params": params, "name": job_name, "experiment": {"name": "Default"}, "isDistributed": False}
        )
        Job(
            logger=logger,
            preload_executor=preload_executor,
            preload_models=preload_models,
            preload_dataset_loaders=preload_dataset_loaders,
        ).run_job(job)
        logger.info("Job completed successfully")
        job_result = DebugJobLogContext().get_result()
        print(job_result)
        return job_result


def _get_init_parameters_from_dict(params):
    """Get __init__ params from list of method_params."""
    for param in params:
        if param["methodName"] == ModelMethodName.init.value:
            return json.loads(param["methodParams"])
    return {}


def local_prune(cache: bool = True, local_registry: bool = True):
    """
    Сlear the cache and local registry directories.

    Parameters
    ----------
    cache:bool=True
        Flag for clearing the cache.

    local_registry:bool=True
        Flag for clearing the local registry.


    Returns
    -------
    None
    """
    if local_registry and os.path.exists(LOCAL_REGISTRY_PATH):
        shutil.rmtree(LOCAL_REGISTRY_PATH)

    if cache and os.path.exists(DEBUG_REGISTRY_PATH):
        shutil.rmtree(DEBUG_REGISTRY_PATH)



from typing import Literal, Optional, Union

from sgqlc.operation import Operation

from ML_management.graphql import schema
from ML_management.graphql.schema import Experiment, ExperimentFilterSettings, TagFilterSettings, UpdateExperimentForm
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.visibility_options import VisibilityOptions


def set_experiment_description(experiment_id: int, description: str) -> Experiment:
    """
    Set experiment description.

    Parameters
    ----------
    experiment_id: int
        Id of an experiment.
    description: str
        Description of an experiment.

    Returns
    -------
    Experiment
        Instance of a experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_description = op.update_experiment(
        experiment_id=experiment_id, update_experiment_form=UpdateExperimentForm(new_description=description)
    )
    set_description.name()
    set_description.description()
    set_description.visibility()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.update_experiment


def set_experiment_visibility(
    experiment_id: int,
    visibility: Union[Literal["private", "public"], VisibilityOptions],
) -> Experiment:
    """
    Set experiment visibility.

    Parameters
    ----------
    experiment_id: int
        Id of an experiment.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        visibility of an experiment.

    Returns
    -------
    Experiment
        Instance of a experiment with meta information.
    """
    visibility = VisibilityOptions(visibility)
    op = Operation(schema.Mutation)
    set_visibility = op.update_experiment(
        experiment_id=experiment_id, update_experiment_form=UpdateExperimentForm(new_visibility=visibility.name)
    )
    set_visibility.name()
    set_visibility.description()
    set_visibility.visibility()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.update_experiment


def set_experiment_tags(experiment_id: int, key: str, values: list[str]) -> Experiment:
    """
    Set experiment tag.

    Parameters
    ----------
    experiment_id: int
        Id of an experiment.
    key: str
        Key of a tag.
    values: list[str]
        Value of a tag.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_experiment_tags = op.set_experiment_tags(experiment_id=experiment_id, key=key, values=values)
    set_experiment_tags.name()
    set_experiment_tags.tags()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.set_experiment_tags


def reset_experiment_tags(experiment_id: int, key: str, values: list[str], new_key: Optional[str] = None) -> Experiment:
    """
    Reset experiment tag.

    Parameters
    ----------
    experiment_id: int
        Id of an experiment.
    key: str
        Key of a tag.
    values: list[str]
        Value of a tag.
    new_key: Optional[str] = None
        New key of a tag.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_experiment_tags = op.reset_experiment_tags(experiment_id=experiment_id, key=key, values=values, new_key=new_key)
    set_experiment_tags.name()
    set_experiment_tags.tags()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.reset_experiment_tags


def delete_experiment_tag(experiment_id: int, key: str, value: Optional[str] = None) -> Experiment:
    """
    Delete experiment tag.

    Parameters
    ----------
    experiment_id: int
        Id of an experiment.
    key: str
        Key of a tag to delete.
    value: Optional[str]=None
        value tag.
    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Mutation)
    set_experiment_tags = op.delete_experiment_tag(experiment_id=experiment_id, key=key, value=value)
    set_experiment_tags.name()
    set_experiment_tags.tags()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.delete_experiment_tag


def get_experiment_by_name(experiment_name: str) -> Experiment:
    """
    Get experiment by its name.

    Parameters
    ----------
    experiment_name: str
        Name of an experiment.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Query)
    experiment_from_name = op.experiment_from_name(name=experiment_name)
    experiment_from_name.name()
    experiment_from_name.tags()
    experiment_from_name.description()
    experiment_from_name.experiment_id()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.experiment_from_name


def get_experiment_by_id(experiment_id: int) -> Experiment:
    """
    Get experiment by it's id.

    Parameters
    ----------
    experiment_id: int
        Id of an experiment.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Query)
    experiment_from_name = op.experiment_from_id(experiment_id=experiment_id)
    experiment_from_name.name()
    experiment_from_name.tags()
    experiment_from_name.description()
    experiment_from_name.experiment_id()

    experiment = send_graphql_request(op=op, json_response=False)
    return experiment.experiment_from_id


def create_experiment(
    experiment_name: str,
    experiment_description: str = "",
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
) -> Experiment:
    """
    Create experiment with given name and description.

    Parameters
    ----------
    experiment_name: str
        Name of the experiment.
    experiment_description: str = ""
        Description of the experiment. Default: "".
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of experiment.  Default: PRIVATE.

    Returns
    -------
    Experiment
        Instance of an experiment with meta information.
    """
    op = Operation(schema.Mutation)
    create_experiment = op.create_experiment(
        experiment_name=experiment_name,
        experiment_description=experiment_description,
        visibility=VisibilityOptions(visibility).name,
    )
    create_experiment.name()
    create_experiment.description()

    experiment = send_graphql_request(op=op, json_response=False)

    return experiment.create_experiment


def pagination_experiment(
    name: Optional[str] = None,
    experiment_id: Optional[int] = None,
    description: Optional[str] = None,
    visibility: Optional[Union[Literal["private", "public"], VisibilityOptions]] = None,
    owner_ids: Optional[list[str]] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> list[Experiment]:
    """
    Search experiment.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the executor.
    experiment_id: Optional[int]=None
        Id of an experiment.
    description: Optional[str]=None
        Description of the experiment.
    visibility: Optional[Union[Literal['private', 'public'], VisibilityOptions]]=None
        Visibility of experiment.
    owner_ids: Optional[list[str]]=None
        Ids of the experiment owner.
    tag_key: Optional[str]=None
        Key of the experiment tag.
    tag_value: Optional[str]=None
        Value of the experiment tag.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[Experiment]
        List of Experiment instance with meta information.
    """
    op = Operation(schema.Query)
    base_query = op.pagination_experiment(
        limit=limit,
        offset=offset,
        filter_settings=ExperimentFilterSettings(
            name=name,
            experiment_id=experiment_id,
            visibility=VisibilityOptions(visibility).name if visibility else None,
            owner_ids=owner_ids,
            description=description,
            tag=TagFilterSettings(key=tag_key, value=tag_value),
        ),
    ).list_experiment
    base_query.name()
    base_query.tags()
    base_query.description()
    base_query.experiment_id()
    base_query.visibility()
    return send_graphql_request(op, json_response=False).pagination_experiment.list_experiment



from enum import Enum
from typing import List

from pydantic import BaseModel


class JobType(str, Enum):
    """Enum job type."""

    build = "build"
    execution = "execution"
    venv = "venv"


class ResourceDataType(str, Enum):
    """Enum resource data type."""

    ROW = "row"
    PERCENT = "percent"


class ResourcePoint(BaseModel):
    """Description of the resource usage at the point."""

    value: float
    timestamp: float


class ResourcesInfo(BaseModel):
    """Specific resource usage."""

    resource_name: str
    points: List[ResourcePoint]
    data_type: ResourceDataType


class ResourceNodeInfo(BaseModel):
    """Node resource usage."""

    node_name: str
    resources: List[ResourcesInfo]



import importlib
import json
import os
import os.path
import subprocess
import sys
import tarfile
import tempfile
import threading
import traceback
from pathlib import Path
from typing import Literal, Optional, Tuple, Union

import yaml
from sgqlc.operation import Operation
from tqdm.autonotebook import tqdm

from ML_management import variables
from ML_management.base_exceptions import MLMClientError
from ML_management.graphql import schema
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.local_debug.debug_job_result import DebugJobLogContext
from ML_management.mlmanagement.log_api import _raise_error
from ML_management.mlmanagement.metainfo import LoadedObject, ObjectMetaInfo
from ML_management.mlmanagement.model_type import ModelType
from ML_management.session import AuthSession
from ML_management.variables import (
    CACHED_LIST_FILENAME,
    CONFIG_KEY_ARTIFACTS,
    DATA,
    FLAVOR_NAME,
    JOB_ARTIFACT_DIRNAME,
    LEGACY_DATA,
    LOCAL_REGISTRY_PATH,
    MLCONFIG,
    MLMODEL_FILE_NAME,
    get_log_service_url,
)


def download_artifacts_by_name_version(
    name: str,
    version: Optional[int],
    path: str,
    dst_path: Optional[str] = None,
    model_type: Union[Literal["model", "executor", "dataset_loader"], ModelType] = ModelType.MODEL,
    verbose: bool = True,
) -> str:
    """Download an artifact by name and version to a local directory, and return a local path for it.

    Parameters
    ==========
    name: str
        Name of the entity.
    version: Optional[int] = None
        Version of the entity. Default: None, "latest" version is used.
    model_type: Union[Literal['model', 'executor', 'dataset_loader'] , ModelType]
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    path: str = ""
        Specific path for artifacts download. Default: "", all artifacts will be downloaded.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    Returns
    =======
    str
        Local path to the entity folder.
    """
    model_type = ModelType(model_type)
    if variables.NO_CACHE:
        local_path = CachedLoader.download_artifacts_by_name_version(
            name=name, version=version, model_type=model_type, path=path, dst_path=dst_path, verbose=verbose
        )
    else:
        local_path = CachedLoader().cached_download_artifacts_by_name_version(
            name=name, version=version, model_type=model_type, path=path, dst_path=dst_path, verbose=verbose
        )

    return local_path


def download_job_artifacts(job_id: int, path: str = "", dst_path: Optional[str] = None, verbose: bool = True) -> str:
    """Download an artifact file or directory from a job to a local directory, and return a local path for it.

    Parameters
    ==========
    job_id: int
        Id of the job.
    path: str = ""
        Specific path for artifacts download. Default: "", all artifacts will be downloaded.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    Returns
    =======
    str
        Local path to artifacts.
    """
    if variables.NO_CACHE:
        local_path = CachedLoader.download_job_artifacts(job_id, path, dst_path, verbose=verbose)
    else:
        local_path = CachedLoader().cached_download_job_artifacts(job_id, path, dst_path, verbose=verbose)
    return local_path


def download_job_metrics(job_id: int, dst_path: Optional[str] = None, verbose: bool = True) -> str:
    """Download  directory of metrics from a job to a local directory, and return a local path for it.

    Parameters
    ==========
    job_id: int
        Id of the job.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    Returns
    =======
    str
        Local path to metrics.
    """
    url = get_log_service_url("download_job_metrics")
    params = {"job_id": job_id}
    return _request_download_artifacts(url, params, dst_path, f"job_{job_id}_metrics", verbose=verbose)


def _load_model_type(
    name: str,
    version: Optional[int],
    model_type: ModelType,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init=None,
    verbose: bool = True,
) -> LoadedObject:
    """Load model from local path."""
    if variables.NO_CACHE:
        local_path = CachedLoader.download_artifacts_by_name_version(
            name=name, version=version, model_type=model_type, path="", dst_path=dst_path, verbose=verbose
        )
    else:
        local_path = CachedLoader().cached_download_artifacts_by_name_version(
            name=name, version=version, model_type=model_type, path="", dst_path=dst_path, verbose=verbose
        )
    if install_requirements:
        _set_model_version_requirements(local_path)
    loaded_model = _load_model_src(local_path, kwargs_for_init)
    with open(os.path.join(local_path, MLCONFIG)) as f:
        config = yaml.safe_load(f)

    metainfo = ObjectMetaInfo(
        name=config["name"],
        version=config.get("version", 0),
        hash_artifacts=config.get("hash_artifacts", "unknown"),
        model_type=model_type,
    )
    loaded_object = LoadedObject(loaded_object=loaded_model, local_path=local_path, metainfo=metainfo)

    return loaded_object


def load_dataset(
    name: str,
    version: Optional[int] = None,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init: Optional[dict] = None,
    verbose: bool = True,
) -> LoadedObject:
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the dataset.
    version: Optional[int] = None
        Version of the dataset. Default: None, "latest" version is used.
    install_requirements: bool = False
        Whether to install dataset requirements. Default: False.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    kwargs_for_init: Optional[dict]: None
        Kwargs for __init__ function of dataset loader when it would be loaded.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    Returns
    =======
    LoadedObject
        The object of the dataset to use.
    """
    return _load_model_type(
        name, version, ModelType.DATASET_LOADER, install_requirements, dst_path, kwargs_for_init, verbose
    )


def _set_model_version_requirements(local_path) -> None:
    """Installing requirements of the model locally."""
    with open(os.path.join(local_path, "requirements.txt")) as req:
        requirements = list(
            filter(
                lambda x: "ml-management" not in x.lower() and "mlflow" not in x.lower() and len(x),
                req.read().split("\n"),
            )
        )
    try:
        if requirements:
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--no-cache-dir",
                    "--default-timeout=100",
                    *requirements,
                ]
            )

    except Exception:
        print(traceback.format_exc())


def load_model(
    name: str,
    version: Optional[int] = None,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init=None,
    verbose: bool = True,
) -> LoadedObject:
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.
    install_requirements: bool = False
        Whether to install model requirements. Default: False.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    kwargs_for_init: Optional[dict]: None
        Kwargs for __init__ function of model when it would be loaded.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    Returns
    =======
    LoadedObject
        The object of the model to use.
    """
    return _load_model_type(name, version, ModelType.MODEL, install_requirements, dst_path, kwargs_for_init, verbose)


def load_executor(
    name: str,
    version: Optional[int] = None,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    verbose: bool = True,
) -> LoadedObject:
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    install_requirements: bool = False
        Whether to install executor requirements. Default: False.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    Returns
    =======
    LoadedObject
        The object of the executor to use.
    """
    return _load_model_type(name, version, ModelType.EXECUTOR, install_requirements, dst_path, verbose=verbose)


def _untar_folder(buff, to_folder):
    try:
        with tarfile.open(mode="r|", fileobj=buff) as tar:
            tar.extractall(to_folder)
    except Exception as err:
        raise MLMClientError("Some error during untar the content.") from err


def _request_download_artifacts(
    url, params: dict, dst_path: Optional[str] = None, extra_dst_path: str = "", verbose: bool = True
):
    path = params.get("path", "")
    with AuthSession().get(url=url, params=params, stream=True) as response:
        _raise_error(response)
        untar = response.headers.get("untar") == "True"
        if dst_path is None:
            dst_path = tempfile.mkdtemp()
        dst_path = os.path.abspath(os.path.normpath(dst_path))
        local_path = os.path.normpath(os.path.join(dst_path, extra_dst_path, os.path.normpath(path)))
        total_size = int(response.headers.get("x-total-size"))

        with tqdm(
            total=total_size,
            disable=not verbose,
            unit_scale=True,
            unit_divisor=1024,
            unit="File",
        ) as pbar:
            if untar:
                r, w = os.pipe()
                with open(r, "rb") as buff:
                    try:
                        thread = threading.Thread(target=_untar_folder, args=(buff, local_path))
                        thread.start()
                    except Exception as err:
                        os.close(r)
                        os.close(w)
                        raise err

                    with open(w, "wb") as wfd:
                        for chunk in response.iter_raw():
                            wfd.write(chunk)
                            pbar.update(len(chunk))
                    thread.join()
                    return local_path
            else:
                dirs = os.path.dirname(local_path)
                if not os.path.exists(dirs):
                    os.makedirs(dirs)
                with open(local_path, "wb") as f:
                    for chunk in response.iter_bytes():
                        f.write(chunk)
                        pbar.update(len(chunk))
                return local_path


def _load_model_src(local_path: str, kwargs_for_init: Optional[dict]):
    if not kwargs_for_init:
        kwargs_for_init = {}
    config_path = os.path.join(local_path, MLCONFIG)
    if os.path.exists(config_path):
        with open(config_path) as file:
            conf = yaml.safe_load(file)

        load_model_path = os.path.join(local_path, conf["load_model_path"])

    elif os.path.exists(os.path.join(local_path, MLMODEL_FILE_NAME)):
        with open(os.path.join(local_path, MLMODEL_FILE_NAME)) as file:
            conf = yaml.safe_load(file)

        # legacy data entity code path for backward compatibility
        code_data_path = (
            conf["flavors"][FLAVOR_NAME][DATA]
            if DATA in conf["flavors"][FLAVOR_NAME]
            else conf["flavors"][FLAVOR_NAME][LEGACY_DATA]
        )
        load_model_path = os.path.join(local_path, code_data_path)

    else:
        raise RuntimeError("MLConfig does not exist.")

    from ML_management.mlmanagement.utils import INIT_FUNCTION_NAME  # circular import

    parts = Path(load_model_path).parts
    extra_sys_path = str(Path(*parts[:-2]))
    module_for_importlib = ".".join(parts[-2:])

    if extra_sys_path not in sys.path:
        sys.path.append(extra_sys_path)
        if DebugJobLogContext().job_name:
            DebugJobLogContext().extra_sys_path.append(extra_sys_path)
    python_model = getattr(importlib.import_module(module_for_importlib), INIT_FUNCTION_NAME)(**kwargs_for_init)
    artifacts = Path(load_model_path) / CONFIG_KEY_ARTIFACTS
    if not artifacts.exists():
        artifacts.mkdir()

    python_model.artifacts = str(artifacts)

    return python_model


def _load_from_src(model_path: str, kwargs_for_init: Optional[dict] = None):
    if not kwargs_for_init:
        kwargs_for_init = {}

    model_path = os.path.abspath(model_path)
    parts = Path(model_path).parts
    extra_sys_path = str(Path(*parts[:-2]))
    module_for_importlib = ".".join(parts[-2:])
    sys.path.append(extra_sys_path)
    if DebugJobLogContext().job_name:
        DebugJobLogContext().extra_sys_path.append(extra_sys_path)
    from ML_management.mlmanagement.utils import INIT_FUNCTION_NAME  # circular import

    python_model = getattr(importlib.import_module(module_for_importlib), INIT_FUNCTION_NAME)(**kwargs_for_init)
    artifacts = Path(model_path) / CONFIG_KEY_ARTIFACTS
    if not artifacts.exists():
        artifacts.mkdir()

    python_model.artifacts = str(artifacts)

    return python_model


class CachedLoader:
    entity_key_pattern: str = "{model_type}:{name}:{version}"

    def __init__(self, registry_path: str = LOCAL_REGISTRY_PATH):
        self.registry_path = registry_path
        os.makedirs(registry_path, exist_ok=True)
        for type_entity in [ModelType.EXECUTOR.value, ModelType.MODEL.value, ModelType.DATASET_LOADER.value]:
            cache_path = os.path.join(registry_path, type_entity)
            os.makedirs(cache_path, exist_ok=True)
            list_path = os.path.join(cache_path, CACHED_LIST_FILENAME)
            if not os.path.exists(list_path):
                with open(list_path, "w") as f:
                    json.dump({}, f)

        job_artifacts = os.path.join(registry_path, JOB_ARTIFACT_DIRNAME)
        os.makedirs(job_artifacts, exist_ok=True)
        list_path = os.path.join(job_artifacts, CACHED_LIST_FILENAME)
        if not os.path.exists(list_path):
            with open(list_path, "w") as f:
                json.dump({}, f)

    @staticmethod
    def download_artifacts_by_name_version(
        name: str,
        version: Optional[int],
        model_type: ModelType,
        path: str,
        dst_path: Optional[str] = None,
        verbose: bool = True,
    ) -> str:
        url = get_log_service_url("download_artifacts_by_name_version")
        params = {
            "path": os.path.normpath(path) if path else path,
            "name": name,
            "model_type": model_type.value,
        }
        if version:
            params["version"] = version

        local_path = _request_download_artifacts(url, params, dst_path, verbose=verbose)
        return local_path

    def cached_download_artifacts_by_name_version(
        self,
        name: str,
        version: Optional[int],
        model_type: ModelType,
        path: str,
        dst_path: Optional[str] = None,
        verbose: bool = True,
    ) -> str:
        list_path = os.path.join(self.registry_path, model_type.value, CACHED_LIST_FILENAME)
        path = os.path.normpath(path) if path else path
        if not version:
            version = self._get_latest_object_version(name, model_type)

        key = self.entity_key_pattern.format(model_type=model_type, name=name, version=version)

        if dst_path is None:
            dst_path = os.path.join(self.registry_path, model_type.value, name, str(version))
            expected_cache_path = os.path.join(dst_path, path)
            if self._check_exist_cache(list_path, key, path) and os.path.exists(expected_cache_path):
                return os.path.abspath(expected_cache_path)

        local_path = self.download_artifacts_by_name_version(
            name=name, version=version, model_type=model_type, path=path, dst_path=dst_path, verbose=verbose
        )
        self._add_list_cache(list_path, key, path)
        return local_path

    @staticmethod
    def download_job_artifacts(
        job_id: int, path: str = "", dst_path: Optional[str] = None, verbose: bool = True
    ) -> str:
        url = get_log_service_url("download_job_artifacts")
        params = {"path": os.path.normpath(path) if path else path, "job_id": job_id}
        local_path = _request_download_artifacts(url, params, dst_path, verbose=verbose)
        return local_path

    def cached_download_job_artifacts(
        self, job_id: int, path: str = "", dst_path: Optional[str] = None, verbose: bool = True
    ) -> str:
        list_path = os.path.join(self.registry_path, JOB_ARTIFACT_DIRNAME, CACHED_LIST_FILENAME)
        path = os.path.normpath(path) if path else path
        if dst_path is None:
            dst_path = os.path.join(self.registry_path, JOB_ARTIFACT_DIRNAME, str(job_id))
            expected_cache_path = os.path.join(dst_path, path)

            if self._check_exist_cache(list_path, job_id, path) and os.path.exists(expected_cache_path):
                return os.path.abspath(expected_cache_path)

        local_path = self.download_job_artifacts(job_id, path, dst_path, verbose=verbose)
        self._add_list_cache(list_path, job_id, path)
        return local_path

    @staticmethod
    def _add_list_cache(list_path, key, path):
        with open(list_path) as f:
            entities = json.load(f)
        if path == "":
            entities[key] = [path]
        elif "" not in entities.get(key, []):
            entities[key] = entities.get(key, []) + [path]
        with open(list_path, "w") as f:
            json.dump(entities, f, indent=4)

    @staticmethod
    def _check_exist_cache(list_path: str, key: Union[int, Tuple[str, str]], path: str) -> bool:
        with open(list_path) as f:
            jobs = json.load(f)
        request_cached = jobs.get(key, [])
        if "" in request_cached:
            return True
        result = list(
            filter(
                lambda p: (Path(path).is_relative_to(Path(p))),
                request_cached,
            )
        )
        return bool(result)

    @staticmethod
    def _get_latest_object_version(name: str, model_type: ModelType) -> int:
        op = Operation(schema.Query)
        _version = schema.ObjectVersionOptionalInput(name=name, version=None)
        base_query = getattr(op, f"{model_type.value}_version_from_name_version")(
            **{f"{model_type.value}_version": _version}
        )
        base_query.version()
        version = send_graphql_request(op, json_response=False)
        return getattr(version, f"{model_type.value}_version_from_name_version").version



from sgqlc.operation import Operation

from ML_management.graphql.schema import schema
from ML_management.graphql.send_graphql_request import send_graphql_request


def list_metric_jobs(metric_name: str, job_ids: list[int]) -> list[dict]:
    """
    Return metrics by job ids and metric_name.

    Parameters
    ----------
    metric_name: str
        Name of the metric.
    job_ids: list[int]
        List of job ids.

    Returns
    -------
    list[dict]
        List of metrics.
    """
    op = Operation(schema.Query)

    op.list_metric_jobs(metric_name=metric_name, job_ids=job_ids)

    metrics = send_graphql_request(op, json_response=False)

    return metrics.list_metric_jobs



import csv
import os
from typing import Dict

from ML_management.variables import DEBUG_REGISTRY_PATH


class LocalRegistry:
    def __init__(self, registry_path: str = DEBUG_REGISTRY_PATH):
        self.registry_path = registry_path

    def get_metrics(self, job_id: int) -> Dict[str, list]:
        job_dir = os.path.join(self.registry_path, "jobs", str(job_id))
        if not os.path.isdir(job_dir):
            raise Exception(f"Job: {job_id} not found in local debug registry.")
        metrics_dir = os.path.join(job_dir, "metrics")
        if not os.path.isdir(metrics_dir):
            return {}
        result = {}
        for filename in os.listdir(metrics_dir):
            with open(os.path.join(metrics_dir, filename)) as file:
                reader = csv.DictReader(file)
                result[os.path.splitext(filename)[0]] = list(reader)

        return result



import importlib
import io
import json
import math
import os
import sys
import tarfile
import threading
import time
import warnings
from contextlib import _GeneratorContextManager
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

import httpx
import yaml
from sgqlc.operation import Operation
from tqdm.autonotebook import tqdm

from ML_management import variables
from ML_management.base_exceptions import *  # noqa: F403
from ML_management.base_exceptions import MLMClientError, MLMServerError
from ML_management.graphql.schema import MetricInput, ParamInput, schema
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.jsonschema_inference import infer_jsonschema
from ML_management.local_debug.local_logger import LocalLogger
from ML_management.mlmanagement.backend_api import get_debug
from ML_management.mlmanagement.batcher import Batcher
from ML_management.mlmanagement.metainfo import ObjectMetaInfo
from ML_management.mlmanagement.metric_autostepper import MetricAutostepper
from ML_management.mlmanagement.model_type import ModelType
from ML_management.mlmanagement.server_mlmanager_exceptions import *  # noqa: F403
from ML_management.mlmanagement.server_mlmanager_exceptions import (
    AuthError,
    InvalidExperimentNameError,
    ModelTypeIsNotFoundError,
)
from ML_management.mlmanagement.utils import (
    INIT_FUNCTION_NAME,
    calculate_hash_directory,
    calculate_size,
    is_model_name_valid,
    validate_predict_config,
)
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.registry.exceptions import *  # noqa: F403
from ML_management.session import AuthSession
from ML_management.variables import (
    CONDA_SIZE_LIMIT,
    CONFIG_KEY_ARTIFACTS,
    DATA,
    DEFAULT_EXPERIMENT,
    EXPERIMENT_NAME_FOR_DATASET_LOADER,
    EXPERIMENT_NAME_FOR_EXECUTOR,
    FILENAME_FOR_INFERENCE_CONFIG,
    INFERENCE_CONFIG_LIMIT,
    get_log_service_url,
)


def _log_object_src(
    artifact_path,
    model_path: str,
    description: str,
    experiment_name: str,
    model_type: ModelType = ModelType.MODEL,
    model_version_tags: Optional[Dict[str, Union[str, list]]] = None,
    registered_model_name: str = "default_name",
    source_model_name=None,
    source_model_version=None,
    source_executor_name=None,
    source_executor_version=None,
    source_executor_role=None,
    upload_model_mode=None,
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
    start_build: bool = True,
    create_venv_pack: bool = False,
    additional_local_packages: Optional[Union[List[str], str]] = None,
    force: bool = False,
    verbose: bool = True,
) -> ObjectMetaInfo:
    """
    Log a src model with custom inference logic and optional data dependencies as an artifact.

    Current run is using.
    Parameter registered_model_name must be not empty string,
            consist of alphanumeric characters, '_'
            and must start and end with an alphanumeric character.
            Validation regexp: "(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+"
    You cannot specify the parameters: loader_module, data_path and the parameters: python_model, artifacts together.
    """
    from ML_management.mlmanagement.upload_model_mode import UploadModelMode  # circular import

    visibility = VisibilityOptions(visibility)

    if not os.path.exists(model_path):
        raise FileNotFoundError(f"Path: {model_path} does not exist.")

    if os.path.basename(model_path) == DATA:
        raise RuntimeError(f"The folder should not be named '{DATA}'. Please rename.") from None

    conda_path = os.path.join(model_path, "conda.yaml")

    if not os.path.exists(conda_path):
        raise FileNotFoundError("There is no conda.yaml file.")

    if os.path.getsize(conda_path) > CONDA_SIZE_LIMIT:
        raise RuntimeError("conda.yaml is too big.")

    inference_config_path = os.path.join(model_path, CONFIG_KEY_ARTIFACTS, FILENAME_FOR_INFERENCE_CONFIG)
    if os.path.isfile(inference_config_path) and os.path.getsize(inference_config_path) > INFERENCE_CONFIG_LIMIT:
        raise RuntimeError(f"{FILENAME_FOR_INFERENCE_CONFIG} is too big.")

    try:
        with open(conda_path) as conda_file:
            dependencies = yaml.safe_load(conda_file)["dependencies"]  # noqa: F841
    except Exception:
        raise RuntimeError("conda.yaml not valid.") from None

    if not description or not isinstance(description, str):
        raise RuntimeError("Please fill in the description.") from None

    if not is_model_name_valid(registered_model_name):
        raise RuntimeError(
            "Parameter 'registered_model_name' must be not empty string, "
            "consist of alphanumeric characters, '_' "
            "and must start and end with an alphanumeric character."
            "Validation regexp: '(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+'"
        )
    del UploadModelMode  # need to delete this because it is not JSON serializable
    if create_venv_pack:
        validate_predict_config(path=os.path.join(f"{model_path}", "artifacts", f"{FILENAME_FOR_INFERENCE_CONFIG}"))

    if model_version_tags:
        for key, value in model_version_tags.items():
            if not isinstance(value, list):
                model_version_tags[key] = [value]
    hash_artifacts = calculate_hash_directory(model_path)

    if get_debug():
        return LocalLogger().log_model(
            registered_model_name,
            model_type,
            model_path,
            hash_artifacts,
        )

    if not force:
        versions_hash = _get_version_hash(registered_model_name, model_type, hash_artifacts, visibility)
        if versions_hash:
            warnings.warn(f"You are trying to upload a copy of an existing versions: {versions_hash}")
            return ObjectMetaInfo(
                name=versions_hash[0].name,
                version=versions_hash[0].version,
                model_type=model_type,
                hash_artifacts=hash_artifacts,
            )

    old_python_path = sys.path.copy()
    old_sys_modules = sys.modules.copy()
    try:
        model_path = os.path.abspath(model_path)
        parts = Path(model_path).parts
        extra_sys_path = str(Path(*parts[:-1]))
        module_for_importlib = ".".join(parts[-1:])
        if additional_local_packages:
            extra_sys_path = str(Path(*parts[:-2]))
            module_for_importlib = ".".join(parts[-2:])
        sys.path.append(extra_sys_path)
        get_object_func = getattr(importlib.import_module(module_for_importlib), INIT_FUNCTION_NAME)
        python_model = get_object_func()  # noqa: F841
        kwargs = {
            "artifact_path": artifact_path,
            "model_path": model_path,
            "description": description,
            "model_version_tags": model_version_tags,
            "python_model": python_model,
            "registered_model_name": registered_model_name,
            "source_model_name": source_model_name,
            "source_model_version": source_model_version,
            "source_executor_name": source_executor_name,
            "source_executor_version": source_executor_version,
            "source_executor_role": source_executor_role,
            "upload_model_mode": upload_model_mode,
            "visibility": visibility,
            "start_build": start_build,
            "create_venv_pack": create_venv_pack,
            "additional_local_packages": additional_local_packages,
            "get_object_func": get_object_func,
            "experiment_name": experiment_name,
            "hash_artifacts": hash_artifacts,
        }
        response = _request_log_model(kwargs, verbose)
        _raise_error(response)
        result = response.json()
    except Exception as err:
        raise err
    finally:
        sys.path = old_python_path
        for module in set(sys.modules) - set(old_sys_modules):
            if hasattr(sys.modules[module], "__file__"):
                if sys.modules[module].__file__ and extra_sys_path in sys.modules[module].__file__:
                    sys.modules.pop(module)

    if additional_local_packages:
        name_version = json.loads(response.content.decode("utf-8"))
        if not isinstance(additional_local_packages, list):
            additional_local_packages = [additional_local_packages]
        for package in additional_local_packages:
            response = _request_log_artifacts(
                package, DATA, name_version["name"], int(name_version["version"]), model_type, verbose
            )
            _raise_error(response)

    return ObjectMetaInfo(
        name=result["name"], version=result["version"], model_type=model_type, hash_artifacts=hash_artifacts
    )


def _get_version_hash(name: str, model_type: ModelType, hash_artifacts: str, visibility: VisibilityOptions) -> list:
    query_name = {
        ModelType.MODEL: "get_model_versions_with_hash",
        ModelType.DATASET_LOADER: "get_dataset_loader_versions_with_hash",
        ModelType.EXECUTOR: "get_executor_versions_with_hash",
    }[model_type]

    op = Operation(schema.Query)
    base = getattr(op, query_name)(name=name, hash_artifacts=hash_artifacts, visibility=visibility.name)
    base.name()
    base.version()
    result = send_graphql_request(op=op, json_response=False)
    return getattr(result, query_name)


def log_executor_src(
    model_path: str,
    registered_name: str,
    description: str,
    start_build: bool = False,
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
    additional_local_packages: Optional[Union[List[str], str]] = None,
    force: bool = False,
    verbose: bool = True,
) -> ObjectMetaInfo:
    """
    Upload executor using folder with source code of the executor.

    model_path folder must contain "__init__.py" file with get_object function
    which have to return instance of the executor.
    Also model_path must contain "conda.yaml" file with dependencies of the executor
    model_path can optionally contain "artifacts" folder with files/folders,
    which would be automatically uploaded with executor
    and path to them can be accessed with self.artifacts within executor class.

    Parameters
    ----------
    model_path: str
        Path to folder with executor.
    registered_name: str
        Name of the executor.
    description: str
        Description of the executor
    start_build: bool
        Is to start build image of the executor right after it was logged. Defaults to False.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of this executor to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
        Defaults to PRIVATE.
    additional_local_packages: Optional[Union[List[str], str]] = None
        Path or list of paths to folder with local dependencies which are not within model_path folder.
        Defaults to None.
    force: str
        This option removes the restriction on logging duplicate versions.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.

    Returns
    -------
    ObjectMetaInfo

    """
    return _log_object_src(
        artifact_path="",
        description=description,
        registered_model_name=registered_name,
        start_build=start_build,
        model_path=model_path,
        visibility=visibility,
        additional_local_packages=additional_local_packages,
        model_type=ModelType.EXECUTOR,
        experiment_name=EXPERIMENT_NAME_FOR_EXECUTOR,
        force=force,
        verbose=verbose,
    )


def log_dataset_loader_src(
    model_path: str,
    registered_name: str,
    description: str,
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
    additional_local_packages: Optional[Union[List[str], str]] = None,
    force: bool = False,
    verbose: bool = True,
) -> ObjectMetaInfo:
    """
    Upload dataset loader using folder with source code of the dataset loader.

    model_path folder must contain "__init__.py" file with get_object function
    which have to return instance of the dataset loader.
    Also model_path must contain "conda.yaml" file with dependencies of the dataset loader
    model_path can optionally contain "artifacts" folder with files/folders,
    which would be automatically uploaded with dataset loader
    and path to them can be accessed with self.artifacts within dataset loader class.

    Parameters
    ----------
    model_path: str
        Path to folder with dataset loader.
    registered_name: str
        Name of the dataset loader.
    description: str
        Description of the dataset loader
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of this dataset loader to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
        Defaults to PRIVATE.
    additional_local_packages: Optional[Union[List[str], str]] = None
        Path or list of paths to folder with local dependencies which are not within model_path folder.
        Defaults to None.
    force: str
        This option removes the restriction on logging duplicate versions.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.

    Returns
    -------
    ObjectMetaInfo
    """
    return _log_object_src(
        artifact_path="",
        description=description,
        registered_model_name=registered_name,
        model_path=model_path,
        visibility=visibility,
        additional_local_packages=additional_local_packages,
        model_type=ModelType.DATASET_LOADER,
        experiment_name=EXPERIMENT_NAME_FOR_DATASET_LOADER,
        force=force,
        verbose=verbose,
    )


def log_model_src(
    model_path: str,
    registered_name: str,
    description: str,
    start_build: bool = True,
    model_version_tags: Optional[Dict[str, Union[str, list]]] = None,
    experiment_name: str = DEFAULT_EXPERIMENT,
    create_venv_pack: bool = False,
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
    additional_local_packages: Optional[Union[List[str], str]] = None,
    force: bool = False,
    verbose: bool = True,
) -> ObjectMetaInfo:
    """
    Log model to remote server.

    Parameters
    ----------
    model_path: str
        Path to the folder with the model files(*model*.py, conda.yaml, __init__.py with get object function, see docs)
    registered_name: str
        The name of the model that will be assigned to the model in the model registry.
        Parameter registered_name should match regexp ``"(([A-Za-z0-9][A-Za-z0-9_]*)?[A-Za-z0-9])+"``.
    description: str
        Description of the model.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of this model to other users. Possible values VisibilityOptions.PRIVATE, PUBLIC.
        Defaults to PRIVATE.
    start_build: bool = True
        Whether to start building the image with the model requirements immediately after uploading.
        This can speed up the subsequent build of the task image. Default: True.
    model_version_tags: Optional[Dict[str, Union[str, list]]] = None
        Define model version tags. Default: None.
    experiment_name: str = "Default"
        Name of the experiment. Default: "Default"
    create_venv_pack: bool = False
        Whether to prepare venv pack for future inference. Default: False.
    additional_local_packages: Optional[Union[List[str], str]] = None
        Path or list of paths to folder with local dependencies which are not within model_path folder.
        Defaults to None.
    force: str
        This option removes the restriction on logging duplicate versions.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.

    Returns
    =======
    ObjectMetaInfo
    """
    return _log_object_src(
        artifact_path="",
        description=description,
        registered_model_name=registered_name,
        model_path=model_path,
        start_build=start_build,
        model_version_tags=model_version_tags,
        create_venv_pack=create_venv_pack,
        visibility=visibility,
        additional_local_packages=additional_local_packages,
        model_type=ModelType.MODEL,
        experiment_name=experiment_name,
        force=force,
        verbose=verbose,
    )


def log_artifact(local_path: str, artifact_path: Optional[str] = None, verbose: bool = True) -> None:
    """
    Log a local file or directory as an artifact of the currently active run.

    If no run is active, this method will create a new active run.

    Parameters
    ==========
    local_path: str
        Path to the file to write.
    artifact_path: Optional[str] = None
        If provided, the directory to write to.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.

    Returns
    =======
    None
    """
    if get_debug():
        LocalLogger().log_artifact(local_path=local_path, artifact_path=artifact_path)
        return
    response = _request_log_artifacts(local_path, artifact_path, verbose=verbose)
    return _raise_error(response)


def _request_log_model(kwargs: dict, verbose: bool = True):
    """
    Send request for log_model function.

    Steps for log model:
    0) Infer jsonschema, raise if it is invalid
    1) open temporary directory
    2) Do init model locally
    3) Pack it to tar file
    4) Send it to server to log model there.
    """
    delete_args_for_save_model_func = [
        "description",
        "model_version_tags",
        "artifact_path",
        "registered_model_name",
        "await_registration_for",
        # now, extra arguments
        "upload_model_mode",
        "source_model_name",
        "source_model_version",
        "visibility",
        "source_executor_name",
        "source_executor_version",
        "source_executor_role",
        "start_build",
        "create_venv_pack",
    ]  # not need for save_model

    extra_imports_args = [
        "submodules",
        "module_name",
        "used_modules_names",
        "extra_modules_names",
        "root_module_name",
        "linter_check",
    ]

    delete_args_for_log_func = [
        "python_model",
        "artifacts",
        "conda_env",
        "pip_requirements",
        "extra_pip_requirements",
        "additional_local_packages",
        "conda_file",
        "dependencies",
        "get_object_func",
    ]  # not need for log model on server

    for delete_arg in extra_imports_args:
        kwargs.pop(delete_arg, None)
    kwargs_for_save_model = kwargs.copy()
    for delete_arg in delete_args_for_save_model_func:
        kwargs_for_save_model.pop(delete_arg, None)
    python_model = kwargs_for_save_model["python_model"]
    get_object_func = kwargs_for_save_model.get("get_object_func")
    # import some modules here because of circular import
    from ML_management.dataset_loader.dataset_loader_pattern import DatasetLoaderPattern
    from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import (
        dataset_loader_pattern_to_methods,
    )
    from ML_management.executor.base_executor import BaseExecutor
    from ML_management.executor.executor_pattern_to_methods_map import executor_pattern_to_methods
    from ML_management.model.model_type_to_methods_map import model_pattern_to_methods
    from ML_management.model.patterns.model_pattern import Model

    experiment_name = kwargs.get("experiment_name", DEFAULT_EXPERIMENT)

    if python_model is not None:
        if isinstance(python_model, Model):
            kwargs["model_type"] = ModelType.MODEL
            model_to_methods = model_pattern_to_methods
            if experiment_name in [
                EXPERIMENT_NAME_FOR_EXECUTOR,
                EXPERIMENT_NAME_FOR_DATASET_LOADER,
            ]:
                raise InvalidExperimentNameError(ModelType.MODEL.value, experiment_name)
        elif isinstance(python_model, BaseExecutor):
            kwargs["model_type"] = ModelType.EXECUTOR
            model_to_methods = executor_pattern_to_methods

            if experiment_name != EXPERIMENT_NAME_FOR_EXECUTOR:
                raise InvalidExperimentNameError(ModelType.EXECUTOR.value, experiment_name)
            # collect all needed model's methods
            kwargs["desired_model_methods"] = python_model.desired_model_methods
            kwargs["desired_dataset_loader_methods"] = python_model.desired_dataset_loader_methods
        elif isinstance(python_model, DatasetLoaderPattern):
            kwargs["model_type"] = ModelType.DATASET_LOADER
            model_to_methods = dataset_loader_pattern_to_methods
            if experiment_name != EXPERIMENT_NAME_FOR_DATASET_LOADER:
                raise InvalidExperimentNameError(kwargs["model_type"].value, experiment_name)
        else:
            raise ModelTypeIsNotFoundError()

        # now we need to infer schemas for methods.
        methods_schema = {}
        for model_type, methods_name_to_schema_map in model_to_methods.items():
            if isinstance(python_model, model_type):
                for method_name_to_schema in methods_name_to_schema_map:
                    model_method = getattr(python_model, method_name_to_schema.value, None)
                    model_method_schema = infer_jsonschema(model_method, get_object_func)
                    methods_schema[method_name_to_schema.value] = model_method_schema

        kwargs["model_method_schemas"] = methods_schema

        for delete_arg in delete_args_for_log_func:
            kwargs.pop(delete_arg, None)

        log_request = {
            "secret_uuid": variables.secret_uuid,
        }

        artifacts_path = os.path.join(kwargs["model_path"], CONFIG_KEY_ARTIFACTS)
        if os.path.isfile(artifacts_path):
            raise Exception(f"The artifact file {artifacts_path} is invalid. The artifact must be a directory.")

        model_folder = kwargs["model_path"]

        del kwargs["model_path"]
        kwargs["loader_module"] = "ML_management.loader.loader"
        log_request["kwargs"] = kwargs
        return _open_pipe_send_request(model_folder, log_request, url=get_log_service_url("log_model"), verbose=verbose)

    else:
        raise Exception("python_model parameter must be specified")


def _request_log_artifacts(
    local_path,
    artifact_path,
    name: Optional[str] = None,
    version: Optional[int] = None,
    model_type: Optional[ModelType] = None,
    verbose: bool = True,
):
    """Send request for log artifact."""
    log_artifact_request = {
        "artifact_path": artifact_path,
        "secret_uuid": variables.secret_uuid,
        "model_type": model_type,
        "name": name,
        "version": version,
    }
    if not os.path.exists(local_path):
        raise FileNotFoundError(f"Path: {local_path} does not exist.")
    if not os.path.isdir(local_path):
        basename = os.path.basename(os.path.normpath(local_path))
        with open(local_path, "rb") as file:
            url = get_log_service_url("log_artifact")

            # upload multipart
            data = {"log_request": json.dumps(log_artifact_request)}
            headers = {"Transfer-Encoding": "chunked"}

            file_content_type = "application/octet-stream"

            files = {"file": (basename, file, file_content_type)}

            with AuthSession().post(
                url=url,
                stream=False,
                data=data,
                files=files,
                headers=headers,
            ) as response:
                return response

    return _open_pipe_send_request(
        local_path, log_artifact_request, url=get_log_service_url("log_artifact"), verbose=verbose
    )


def _raise_error(response: httpx.Response):
    if response.status_code == 500:
        raise MLMServerError("Internal server error.")
    if response.status_code != 200:
        detail = response.read().decode()
        if not detail:
            raise MLMServerError("Internal server error.")
        try:
            detail = json.loads(detail).get("detail")
        except Exception:
            raise MLMServerError(f"Server error '{detail}' with code {response.status_code}") from None
        if not (
            isinstance(detail, dict) and "exception_class" in detail and ("params" in detail or "message" in detail)
        ):
            raise MLMServerError(detail)
        if "params" in detail:
            error = getattr(sys.modules[__name__], detail["exception_class"])(**detail["params"])
        else:
            error = getattr(sys.modules[__name__], detail["exception_class"])(detail["message"])
        if isinstance(error, AuthError):
            error.args = (
                f"{error.args[0]}. "
                "Possible reason: you are trying to upload a version of an object owned by another user.",
            )
        raise error


class ProgressFile(io.RawIOBase):
    def __init__(self, file_obj, pbar):
        self.file_obj = file_obj
        self.pbar = pbar

    def read(self, size=-1):
        data = self.file_obj.read(size)
        self.pbar.update(len(data))
        return data


def _open_pipe_send_request(folder, request_dict, url, verbose: bool = True):
    total_size = calculate_size(folder)

    r, w = os.pipe()

    try:
        thread = threading.Thread(target=_tar_folder, args=(w, folder))
        thread.start()
    except Exception as err:
        os.close(r)
        os.close(w)
        raise err
    with tqdm(
        total=total_size,
        disable=not verbose,
        unit_scale=True,
        unit_divisor=1024,
        unit="File",
    ) as pbar:
        with open(r, "rb") as buff:
            progress_file = ProgressFile(buff, pbar)

            with _request(url, request_dict, progress_file, os.path.basename(folder), True) as response:
                return response


def _request(
    url,
    request_dict,
    file=None,
    basename=None,
    is_tar=False,
    stream=False,
) -> _GeneratorContextManager:
    """Create log request and send it to server."""
    if not file:
        return AuthSession().post(url=url, stream=stream, json=request_dict)

    # upload multipart
    data = {"log_request": json.dumps(request_dict)}
    headers = {"Transfer-Encoding": "chunked"}

    file_content_type = "application/octet-stream"
    if is_tar:
        file_content_type = "application/x-tar"

    files = {"file": (basename, file, file_content_type)}

    return AuthSession().post(
        url=url,
        stream=stream,
        data=data,
        files=files,
        headers=headers,
    )


def _tar_folder(w, model_folder):
    try:
        with open(w, "wb") as buff:
            with tarfile.open(mode="w|", fileobj=buff) as tar:
                tar.add(model_folder, arcname=os.path.basename(model_folder))
    except Exception as err:
        raise MLMClientError("Some error during tar the content.") from err


def log_metric(key: str, value: float, step: int = 0, no_wait=False):
    """
    Log a metric under the current job.

    Parameters
    ==========
    key: str
        Metric name (string).
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.

    value: float
        Metric value (float). Note that some special values such as +/-
        Infinity may be replaced by other values depending on the store.
        For example, the SQLAlchemy store replaces +/- Infinity with max / min float values.
        All backend stores will support values up to length 5000, but some may support larger values.

    step: int
        Metric step (int). Defaults to zero if unspecified.

    no_wait: bool
        The no_wait flag allows you to log metrics directly, without using optimization (batching).

    Returns
    =======
    None
    """
    if not variables.secret_uuid:
        raise MLMClientError("The log_metric function must be called from the active job.")
    if not (isinstance(value, (float, int)) and math.isfinite(value)):
        warnings.warn(f"The log_metric function can log only float or integer values. Value {value} is not log.")
        return

    if not isinstance(key, str):
        warnings.warn(
            f"The log_metric function can log only str name metric and float or integer values. "
            f"{key}: {value} is not log."
        )
        return
    timestamp = int(time.time() * 10e6)
    autostep = MetricAutostepper().get_next_step(key)

    metric = {"key": key, "value": value, "step": step, "timestamp": timestamp, "autostep": autostep}
    if get_debug():
        LocalLogger().log_metrics([metric])
        return

    if no_wait:
        op = Operation(schema.Mutation)
        op.log_metric(
            metric=MetricInput(**metric),
            secret_uuid=variables.secret_uuid,
        )
        try:
            return send_graphql_request(op, json_response=False).log_metric
        except Exception as err:
            warnings.warn(str(err))
    else:
        Batcher().log_metrics([MetricInput(key=key, value=value, step=step, timestamp=timestamp, autostep=autostep)])


def log_metrics(metrics: Dict[str, float], step: int = 0, no_wait=False):
    """
    Log a metrics under the current job.

    Parameters
    ==========
    metrics: dict[str, float]
        Key(string): metric name.
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.
        Value(float): metric value. Note that some special values such as +/-
        Infinity may be replaced by other values depending on the store.
        For example, the SQLAlchemy store replaces +/- Infinity with max / min float values.
        All backend stores will support values up to length 5000, but some may support larger values.

    step: int
        Metric step (int). Defaults to zero if unspecified.

    no_wait: bool
        The no_wait flag allows you to log metrics directly, without using optimization (batching).

    Returns
    =======
    None
    """
    if not variables.secret_uuid:
        raise MLMClientError("The log_metric function must be called from the active job.")

    metrics_to_log = {}
    for key, value in metrics.items():
        if not (isinstance(value, (float, int)) and math.isfinite(value)):
            warnings.warn(f"The log_metrics function can log only float or integer values. {key}: {value} is not log.")
            continue
        if not isinstance(key, str):
            warnings.warn(
                f"The log_metrics function can log only str name metric and float or integer values. "
                f"{key}: {value} is not log."
            )
            continue
        metrics_to_log[key] = value
    timestamp = int(time.time() * 10e6)

    metrics = [
        {
            "key": key,
            "value": value,
            "step": step,
            "timestamp": timestamp,
            "autostep": MetricAutostepper().get_next_step(key),
        }
        for key, value in metrics_to_log.items()
    ]

    if get_debug():
        LocalLogger().log_metrics(metrics)
        return
    metrics = [MetricInput(**m) for m in metrics]
    if no_wait:
        op = Operation(schema.Mutation)
        op.log_metrics(
            metrics=metrics,
            secret_uuid=variables.secret_uuid,
        )
        try:
            return send_graphql_request(op, json_response=False).log_metrics
        except Exception as err:
            warnings.warn(str(err))
    else:
        Batcher().log_metrics(metrics)


def _log_param(key: str, value: str):
    """
    Log str parameter (e.g. model hyperparameter) under the current job.

    Parameters
    ==========
    key: str
        Param name (string).
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.

    value: str
        Param value (string).

    Returns
    =======
    None
    """
    if not isinstance(value, str):
        warnings.warn(f"The log_param function can log only string values. Value {value} is not log.")
        return
    if not variables.secret_uuid:
        raise MLMClientError("The log_param function must be called from the active job.")

    if get_debug():
        LocalLogger().log_params({key: value})
        return
    op = Operation(schema.Mutation)
    op.log_param(param=ParamInput(key=key, value=value), secret_uuid=variables.secret_uuid)
    try:
        return send_graphql_request(op, json_response=False).log_param
    except Exception as err:
        warnings.warn(str(err))


def _log_params(params: Dict[str, str]):
    """
    Log a batch of params for the current job.

    Parameters
    ==========
    params: dict
        Key parameter value pairs.
        Кey is the name of the parameter (string).
        This string may only contain alphanumerics, underscores (_), dashes (-),
        periods (.), spaces ( ), and slashes (/).
        All backend stores will support keys up to length 250, but some may support larger keys.
        Param value (string).

    Returns
    =======
    None
    """
    if not variables.secret_uuid:
        raise MLMClientError("The log_metric function must be called from the active job.")

    for key, value in params:
        if not isinstance(value, str):
            warnings.warn(f"The log_params function can log only string values. Value {value} is not logged.")
        params.pop(key)
    if not params:
        return
    if get_debug():
        LocalLogger().log_params(params)
        return

    op = Operation(schema.Mutation)
    op.log_params(
        params=[ParamInput(key=key, value=value) for key, value in params.items()], secret_uuid=variables.secret_uuid
    )
    try:
        return send_graphql_request(op, json_response=False).log_params
    except Exception as err:
        warnings.warn(str(err))



from typing import List, Literal, Optional, Union

import pandas as pd
from sgqlc.operation import Operation

from ML_management.graphql import schema
from ML_management.graphql.schema import (
    ModelInfo,
    ModelVersionInfo,
    ObjectFilterSettings,
    ObjectVersionFilterSettings,
    TagFilterSettings,
    TimestampInterval,
    UpdateObjectForm,
)
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.sdk.sdk import _entity, _to_datetime


def list_model() -> pd.DataFrame:
    """
    List available models.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with list of available models.
    """
    op = Operation(schema.Query)
    _entity(op.list_model)

    json_data = send_graphql_request(op)
    df = pd.DataFrame.from_dict(json_data["listModel"])
    if not df.empty:
        df = _to_datetime(df, ["creationTimestamp", "lastUpdatedTimestamp"])
    return df


def delete_model(model_name: str) -> bool:
    """
    Delete model and all of it's versions.

    Parameters
    ----------
    model_name: str
        Name of the model to delete.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.delete_model(name=model_name)
    return send_graphql_request(op)["deleteModel"]


def set_model_tags(name: str, key: str, values: list[str]) -> ModelInfo:
    """
    Set model tags.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    values: list[str]
        Value tags.

    Returns
    -------
    ModelInfo
        ModelInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_tag = op.set_model_tags(name=name, key=key, values=values)
    _entity(set_tag)
    model = send_graphql_request(op=op, json_response=False).set_model_tags
    return model


def reset_model_tags(name: str, key: str, values: list[str], new_key: Optional[str] = None) -> ModelInfo:
    """
    Reset model tags.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    values: list[str]
        Value tags.
    new_key: Optional[str] = None
        New key of a tag.


    Returns
    -------
    ModelInfo
        ModelInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_tag = op.reset_model_tags(name=name, key=key, values=values, new_key=new_key)
    _entity(set_tag)
    model = send_graphql_request(op=op, json_response=False).reset_model_tags
    return model


def delete_model_tag(name: str, key: str, value: Optional[str] = None) -> ModelInfo:
    """
    Delete model tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.
    Returns
    -------
    ModelInfo
        ModelInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    delete_tag = op.delete_model_tag(name=name, key=key, value=value)
    _entity(delete_tag)
    model = send_graphql_request(op=op, json_response=False).delete_model_tag
    return model


def set_model_description(name: str, description: str) -> ModelInfo:
    """
    Set model description.

    Parameters
    ----------
    name: str
        Name of the model.
    description: str
        Description model.

    Returns
    -------
    ModelInfo
        ModelInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_description = op.update_model(name=name, update_model_form=UpdateObjectForm(new_description=description))
    _entity(set_description)

    model = send_graphql_request(op=op, json_response=False).update_model
    return model


def set_model_visibility(name: str, visibility: Union[Literal["private", "public"], VisibilityOptions]) -> ModelInfo:
    """
    Set model visibility.

    Parameters
    ----------
    name: str
        Name of the model.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        visibility model.

    Returns
    -------
    ModelInfo
        ModelInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_visibility = op.update_model(
        name=name, update_model_form=UpdateObjectForm(new_visibility=VisibilityOptions(visibility).name)
    )
    _entity(set_visibility)

    model = send_graphql_request(op=op, json_response=False).update_model
    return model


def list_model_version(name: str) -> pd.DataFrame:
    """
    List available versions of the model with such name.

    Parameters
    ----------
    name: str
        Name of the model.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with a list of available model versions.
    """
    op = Operation(schema.Query)
    base_query = op.model_from_name(name=name).list_model_version
    base_query.version()
    base_query.creation_timestamp()
    base_query.name()
    json_data = send_graphql_request(op)

    df = pd.DataFrame.from_dict(json_data["modelFromName"]["listModelVersion"])
    df = _to_datetime(df, ["creationTimestamp"])

    return df.sort_values(by=["version"], ignore_index=True)


def get_model_version(name: str, version: Optional[int] = None) -> ModelVersionInfo:
    """
    Meta information about the model version by the model name and version.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.model_version_from_name_version(model_version=_model_version)
    base_query.name()
    base_query.tags()
    base_query.version()
    base_query.experiment.name()
    base_query.experiment.experiment_id()
    base_query.build_job.status()
    base_query.build_job.build_object_name()
    base_query.available_executor_versions.name()
    base_query.available_executor_versions.version()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.model_version_from_name_version


def rebuild_model_version_image(name: str, version: int) -> str:
    """
    Start building new docker image for specified model version.

    Parameters
    ----------
    name: str
        Name of the model.
    version: int
        Version of the model

    Returns
    -------
    str
        name of new docker image for specified model version.
    """
    op = Operation(schema.Mutation)
    _model_version = schema.ObjectVersionInput(name=name, version=version)
    op.rebuild_model_version_image(model_version=_model_version)
    result = send_graphql_request(op=op)
    return result["rebuildModelVersionImage"]


def cancel_venv_build_job_for_model_version(model_name: str, model_version: int) -> bool:
    """
    Cancel running or planned build job of model's environment.

    Parameters
    ----------
    model_name: str
        The name of the model.
    model_version: int
        The version of the model.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.cancel_venv_build_job_for_model_version(name=model_name, version=model_version)
    return send_graphql_request(op)["cancelVenvBuildJobForModelVersion"]


def cancel_build_job_for_model_version(model_name: str, model_version: int) -> bool:
    """
    Cancel running or planned build job of model image.

    Parameters
    ----------
    model_name: str
        The name of the model.
    model_version: int
        The version of the model.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.cancel_build_job_for_model_version(name=model_name, version=model_version)
    return send_graphql_request(op)["cancelBuildJobForModelVersion"]


def delete_model_version(model_name: str, model_version: int):
    """
    Delete version of a model.

    Parameters
    ----------
    model_name: str
        The name of the model.
    model_version: int
        The version of the model.

    Returns
    -------
    None
    """
    op = Operation(schema.Mutation)
    model_version_choice = schema.ObjectVersionInput(name=model_name, version=model_version)
    op.delete_model_version_from_name_version(model_version=model_version_choice)
    send_graphql_request(op, json_response=False)


def get_latest_model_version(name: str) -> ModelVersionInfo:
    """
    Latest model version by the model name.

    Parameters
    ----------
    name: str
        Name of the model.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    return get_model_version(name)


def get_initial_model_version(name: str) -> ModelVersionInfo:
    """
    Initial model version by the model name.

    Parameters
    ----------
    name: str
        Name of the model.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Query)
    version = op.model_from_name(name=name).init_model_version()
    version.name()
    version.version()
    version.tags()
    version.description()
    model_version = send_graphql_request(op, json_response=False).model_from_name.init_model_version
    return model_version


def serve_model(name: str, version: Optional[int] = None, gpu: bool = False) -> str:
    """
    Start model serving in triton service.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.
    gpu: bool
        Determine which device will be used: GPU or CPU. Default: False, CPU is used.

    Returns
    -------
    str:
        endpoint path to make inference requests for model version.
    """
    if version is None:
        version = get_latest_model_version(name).version

    op = Operation(schema.Mutation)
    serving_parameters = schema.ModelServingInput(
        model_version=schema.ObjectVersionInput(name=name, version=version), gpu=gpu
    )
    op.serve_model(serving_parameters=serving_parameters)
    send_graphql_request(op, timeout=None)
    return f"/v2/models/{name}/versions/{version}/infer"


def stop_model_serving(name: str, version: Optional[int] = None) -> None:
    """
    Stop model serving in triton service.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.
    """
    if version is None:
        version = get_latest_model_version(name).version

    op = Operation(schema.Mutation)
    op.stop_model_serving(model_version=schema.ObjectVersionInput(name=name, version=version))
    send_graphql_request(op=op, timeout=None)
    print("Serving has been successfully stopped.")


def check_inference_model_readiness(name: str, version: Optional[int] = None) -> Optional[str]:
    """
    Check if the model is ready to accept requests.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.

    Returns
    -------
    Optional[str]:
        endpoint path to make requests if model is ready else None
    """
    if version is None:
        version = get_latest_model_version(name).version

    op = Operation(schema.Query)
    op.is_inference_model_ready(model_version=schema.ObjectVersionInput(name=name, version=version))
    json_data = send_graphql_request(op=op)
    result = f"/v2/models/{name}/versions/{version}/infer" if json_data["isInferenceModelReady"] else None
    return result


def set_model_version_description(name: str, version: int, description: str) -> ModelVersionInfo:
    """
    Set model version description.

    Parameters
    ----------
    name: str
        Name of the model.
    version: int
        Version of the model.
    description: str
        Description model version.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_description = op.update_model_version(
        model_version=choice, update_model_version_form=UpdateObjectForm(new_description=description)
    )
    set_description.name()
    set_description.version()
    set_description.description()

    model = send_graphql_request(op=op, json_response=False).update_model_version
    return model


def set_model_version_visibility(
    name: str,
    version: int,
    visibility: Union[Literal["private", "public"], VisibilityOptions],
) -> ModelVersionInfo:
    """
    Set model version visibility.

    Parameters
    ----------
    name: str
        Name of the model.
    version: int
        Version of the model.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        visibility model version.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_visibility = op.update_model_version(
        model_version=choice,
        update_model_version_form=UpdateObjectForm(new_visibility=VisibilityOptions(visibility).name),
    )
    set_visibility.name()
    set_visibility.version()
    set_visibility.visibility()

    model = send_graphql_request(op=op, json_response=False).update_model_version
    return model


def set_model_version_tags(name: str, version: int, key: str, values: list[str]) -> ModelVersionInfo:
    """
    Set model version tags.

    Parameters
    ----------
    name: str
        Name of the model.
    version: int
        Version of the model.
    key: str
        Key tag.
    values: list[str]
        Value tag.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_tag = op.set_model_version_tags(model_version=choice, key=key, values=values)
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    model = send_graphql_request(op=op, json_response=False).set_model_version_tags
    return model


def reset_model_version_tags(
    name: str, version: int, key: str, values: list[str], new_key: Optional[str] = None
) -> ModelVersionInfo:
    """
    Reset model version tags.

    Parameters
    ----------
    name: str
        Name of the model.
    version: int
        Version of the model.
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.

    Returns
    -------
    ModelVersionInfo
        ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_tag = op.reset_model_version_tags(model_version=choice, key=key, values=values, new_key=new_key)
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    model = send_graphql_request(op=op, json_response=False).reset_model_version_tags
    return model


def delete_model_version_tag(name: str, version: int, key: str, value: Optional[str] = None) -> ModelVersionInfo:
    """
    Delete model version tag.

    Parameters
    ----------
    name: str
        Name of the model.
    version: int
        Version of the model.
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.

    Returns
    -------
    ModelVersion
        Model version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    delete_tag = op.delete_model_version_tag(model_version=choice, key=key, value=value)
    delete_tag.name()
    delete_tag.version()
    delete_tag.tags()
    model = send_graphql_request(op=op, json_response=False).delete_model_version_tag
    return model


def get_model_version_conda_env(name: str, version: int) -> dict:
    """
    Condas configuration for the model version by the model name and version.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.

    Returns
    -------
    Dict
        Dict with conda configuration.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.model_version_from_name_version(model_version=_model_version)
    base_query.get_conda_env()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.model_version_from_name_version.get_conda_env


def get_model_version_requirements(name: str, version: int) -> list:
    """
    Requirements for the model version by the model name and version.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.

    Returns
    -------
    List
        List of requirements.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.model_version_from_name_version(model_version=_model_version)
    base_query.list_requirements()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.model_version_from_name_version.list_requirements


def pagination_model(
    name: Optional[str] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[Union[Literal["private", "public"], VisibilityOptions]] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[ModelInfo]:
    """
    Search models.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the model.
    tag_key: Optional[str]=None
        Key of the model tag.
    tag_value: Optional[str]=None
        Value of the model tag.
    description: Optional[str]=None
        Description of the model.
    visibility: Optional[Union[Literal['private', 'public'], VisibilityOptions]]=None
        Visibility of model.
    owner_ids: Optional[list[str]]=None
        Ids of the model owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the model.
    creation_to: Optional[int]=None
        Creation timestamp from of the model.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the model.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the model.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[ModelInfo]
        List of ModelInfo instance with meta information.
    """
    visibility = VisibilityOptions(visibility) if visibility else visibility
    op = Operation(schema.Query)
    base_query = op.pagination_model(
        filter_settings=ObjectFilterSettings(
            name=name,
            description=description,
            visibility=visibility,
            owner_ids=owner_ids,
            tag=TagFilterSettings(key=tag_key, value=tag_value),
            creation_interval=TimestampInterval(start=creation_from, end=creation_to),
            last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
        ),
        limit=limit,
        offset=offset,
    ).list_model
    _entity(base_query)

    return send_graphql_request(op, json_response=False).pagination_model.list_model


def pagination_model_version(
    name: str,
    version: Optional[int] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[VisibilityOptions] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[ModelVersionInfo]:
    """
    Search model versions.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the model version.
    version: Optional[int] = None
        Version of the model version.
    tag_key: Optional[str]=None
        Key of the model version tag.
    tag_value: Optional[str]=None
        Value of the model version tag.
    description: Optional[str]=None
        Description of the model version.
    visibility: Optional[str]=None
        Visibility of model version.
    owner_ids: Optional[list[str]]=None
        Ids of the experiment owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the model version.
    creation_to: Optional[int]=None
        Creation timestamp from of the model version.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the model version.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the model version.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[ModelVersionInfo]
        List of ModelVersionInfo instance with meta information.
    """
    op = Operation(schema.Query)
    base_query = (
        op.model_from_name(name=name)
        .pagination_model_version(
            limit=limit,
            offset=offset,
            filter_settings=ObjectVersionFilterSettings(
                version=version,
                description=description,
                visibility=visibility,
                owner_ids=owner_ids,
                tag=TagFilterSettings(key=tag_key, value=tag_value),
                creation_interval=TimestampInterval(start=creation_from, end=creation_to),
                last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
            ),
        )
        .list_model_version
    )
    base_query.name()
    base_query.version()
    base_query.tags()
    base_query.description()
    return send_graphql_request(op, json_response=False).model_from_name.pagination_model_version.list_model_version



import enum
from inspect import getfullargspec

from ML_management.jsonschema_inference.jsonschema_inference import get_function_args


def method_params_prepare(model, method_name: str, params: dict) -> dict:
    prepared = params.copy()
    func = getattr(model, method_name)
    spec = getfullargspec(func)
    all_args, _ = get_function_args(spec, func)
    if "self" in all_args:
        all_args.remove("self")
    for arg in set(params.keys()).intersection(all_args):
        annotation = spec.annotations[arg]
        if hasattr(annotation, "schema"):
            prepared[arg] = annotation.model_validate(params[arg])
        elif issubclass(annotation.__class__, enum.EnumMeta):
            prepared[arg] = annotation(params[arg])
    return prepared



import traceback
from typing import Literal, Optional, Union

from sgqlc.operation import Operation

from ML_management import variables
from ML_management.graphql import schema
from ML_management.graphql.schema import ExecutionJob
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.batcher import Batcher
from ML_management.mlmanagement.metric_autostepper import MetricAutostepper
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.variables import DEFAULT_EXPERIMENT


class ActiveJob:
    """
    A context manager that allows for the execution of a task locally.

    This class provides a convenient way to run a job locally.

    """

    def __init__(self, secret_uuid):
        self.secret_uuid = secret_uuid
        self.job = self._start()
        self.__is_distributed = self.job.is_distributed

    def __enter__(self) -> "ActiveJob":
        return self

    def _start(self) -> ExecutionJob:
        op = Operation(schema.Mutation)
        base_query = op.start_job(secret_uuid=self.secret_uuid)
        base_query.name()
        base_query.id()
        base_query.experiment.name()
        base_query.is_distributed()
        _query_job_params(base_query)
        job = send_graphql_request(op=op, json_response=False).start_job
        variables.secret_uuid = self.secret_uuid
        return job

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__is_distributed:
            Batcher().wait_log_metrics()
            return
        exception_traceback = None
        message = None
        status = "SUCCESSFUL"
        if exc_type:
            exception_traceback = traceback.format_exc()
            message = ": ".join([exc_type.__name__, str(exc_val)])
            status = "FAILED"

        return stop_job(status, message, exception_traceback)


def start_job(
    job_name: Optional[str] = None,
    experiment_name: str = DEFAULT_EXPERIMENT,
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
) -> ActiveJob:
    """
    Create local job.

    Usage::

        with start_job('my-beautiful-job') as job:
            mlmanagement.log_metric(...)
            mlmanagement.log_artifacts(...)


    Or::

        start_job('my-beautiful-job')
        mlmanagement.log_metric(...)
        mlmanagement.log_artifacts(...)
        stop_job()


    Parameters
    ----------
    job_name: str | None=None
        Name of the new job. If not passed, it will be generated.
    experiment_name: str = "Default"
        Name of the experiment. Default: "Default"
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of this job to other users. Default: PRIVATE.

    Returns
    -------
    ActiveJob
        Active job.
    """
    visibility = VisibilityOptions(visibility)
    op = Operation(schema.Mutation)
    op.create_local_job(job_name=job_name, experiment_name=experiment_name, visibility=visibility.name)
    secret_uuid = send_graphql_request(op=op, json_response=False).create_local_job
    return ActiveJob(secret_uuid)


def stop_job(
    status: Literal["SUCCESSFUL", "FAILED"] = "SUCCESSFUL",
    message: Optional[str] = None,
    exception_traceback: Optional[str] = None,
) -> None:
    """
    Stop local job.

    Parameters
    ----------
    status: Literal["SUCCESSFUL", "FAILED"] = "SUCCESSFUL"
        Status of the job. If not passed, it will be SUCCESSFUL.
    message: Optional[str] = None
        Extra message for the job. Default: None
    exception_traceback: Optional[str] = None
        Error traceback of the job. Default: None

    Returns
    -------
    None
    """
    Batcher().wait_log_metrics()
    op = Operation(schema.Mutation)
    op.stop_job(
        secret_uuid=variables.secret_uuid, status=status, message=message, exception_traceback=exception_traceback
    )
    try:
        _ = send_graphql_request(op=op, json_response=False).stop_job
    finally:
        variables.secret_uuid = None
        MetricAutostepper().clear()


def _query_job_params(base_query):
    base_query.params()
    base_query.params.resources.gpu_number()
    base_query.params.models_schemas()
    base_query.params.list_role_model_params()
    base_query.params.list_role_data_params()
    base_query.params.list_role_data_params.data_params()
    base_query.params.list_role_data_params.role()
    base_query.params.list_role_model_params.model_params()
    base_query.params.list_role_model_params.upload_params()
    base_query.params.list_role_model_params.role()
    base_query.params.executor_params()
    base_query.params.executor_params.executor_method_params()
    base_query.params.executor_params.executor_version_choice()



"""Module with session for authenticated requests to server."""
import os
import posixpath
import warnings
from contextlib import contextmanager
from http import HTTPStatus
from typing import Generator

import httpx
from httpx import Response
from sgqlc.endpoint.http import HTTPEndpoint
from sgqlc.operation import Operation
from websocket import WebSocket, create_connection

import ML_management
from ML_management.base_exceptions import MLMClientError
from ML_management.singleton_pattern import Singleton
from ML_management.variables import SERVER_META, get_mlm_credentials, get_server_url


class InvalidCredentialsError(MLMClientError):
    """Exception for invalid login-password pair."""

    pass


class AuthSession(metaclass=Singleton):
    """Extend the standard session functionality with authentication functionality."""

    def __init__(self) -> None:
        self.cookies = {}
        if self._try_set_cookies() or self._try_authenticate_by_credentials():
            self.check_version()

    def check_version(self):
        client_version = ML_management.__version__
        response = httpx.get(posixpath.join(get_server_url(), SERVER_META), cookies=self.cookies)
        try:
            response.raise_for_status()
        except Exception:
            warnings.warn(
                "Can't get ml-management version of the server. "
                f"Server status {response.status_code} with content {response.text}."
            )
            return
        server_version = response.json().get("PLATFORM_VERSION", "unknown")
        if client_version != server_version:
            warnings.warn(
                f"Your ml-management version ({client_version}) does not match the "
                f"version on the server ({server_version})."
            )

    @contextmanager
    def get(self, url: str, stream: bool = False, **kwargs) -> Generator[httpx.Response, None, None]:
        """Proxy get request."""
        with httpx.Client(timeout=None) as client:
            request = client.build_request("GET", url, cookies=self.cookies, **kwargs)
            response = client.send(request=request, stream=stream)
            yield response

        # if token was updated, update our cookie
        self._update_cookies(response, ["kc-access"])

    @contextmanager
    def post(self, url: str, stream: bool = False, **kwargs) -> Generator[httpx.Response, None, None]:
        """Proxy post request."""
        with httpx.Client(timeout=None) as client:
            request = client.build_request("POST", url, cookies=self.cookies, **kwargs)

            # httpx insert content-length even if transfer-encoding was set to 'chunked'
            # so delete content-length header if header transfer-encoding was set
            transfer_encoding_header = request.headers.get("transfer-encoding")
            if transfer_encoding_header == "chunked":
                request.headers.pop("content-length", "")

            response = client.send(request=request, stream=stream)

            yield response

        # if token was updated, update our cookie
        self._update_cookies(response, ["kc-access"])

    # For sdk auth purposes
    def sgqlc_request(self, operation: Operation, timeout=None) -> dict:
        """Make request to /graphql for operation."""
        cookie_header = self._get_cookie_header()
        return HTTPEndpoint(
            posixpath.join(get_server_url(), "graphql"), base_headers={"Cookie": cookie_header}, timeout=timeout
        )(operation)

    def instantiate_websocket_connection(self, url: str) -> WebSocket:
        """Create websocket connection."""
        ws = create_connection(url, cookie=self._get_cookie_header())
        return ws

    def _update_cookies(self, response: Response, cookie_names: list) -> None:
        """Update cookies from cookie_names list."""
        for cookie_name in cookie_names:
            if cookie := response.cookies.get(cookie_name):
                self.cookies[cookie_name] = cookie

    def _get_cookie_header(self) -> str:
        return "; ".join(f"{cookie_name}={cookie_value}" for cookie_name, cookie_value in self.cookies.items())

    def _try_set_cookies(self) -> bool:
        kc_access, kc_state = os.getenv("kc_access"), os.getenv("kc_state")
        if kc_access is not None and kc_state is not None:
            for name, value in zip(["kc-access", "kc-state"], [kc_access, kc_state]):
                self.cookies[name] = value
            return True
        return False

    def _try_authenticate_by_credentials(self) -> bool:
        login, password = get_mlm_credentials()

        try:
            response = httpx.post(
                posixpath.join(get_server_url(), "oauth", "login"), data={"username": login, "password": password}
            )
        except httpx.ConnectError:
            raise RuntimeError("Server is unavailable or server url is incorrect.") from None

        if response.status_code == HTTPStatus.UNAUTHORIZED:
            raise InvalidCredentialsError(f"User with login {login} and password {password} does not exist.")
        if response.status_code != HTTPStatus.OK:
            if not login or not password:
                raise InvalidCredentialsError("You must provide both login and password credentials.")
            response.raise_for_status()
        self.cookies = {
            "kc-state": response.cookies.get("kc-state"),
            "kc-access": response.cookies.get("kc-access"),
        }
        return True



"""DatasetLoader template for poisoned dataset."""
from abc import abstractmethod
from typing import List

from ML_management.dataset_loader.dataset_loader_pattern import DatasetLoaderPattern


class PoisonedImagesDatasetLoader(DatasetLoaderPattern):
    """Implementation of the DatasetLoader for interacting with poisoned data."""

    def __init__(self):
        """Init dataset loader class."""
        super().__init__()

    @abstractmethod
    def get_images_names(self, **kwargs) -> List[str]:
        """Define get_images_names function."""
        raise NotImplementedError

    @abstractmethod
    def get_is_poisoned_by_names(self, names: List[str], **kwargs) -> List[bool]:
        """Define get_is_poisoned_by_names function."""
        raise NotImplementedError

    @abstractmethod
    def get_clean_labels_by_names(self, names: List[str], **kwargs) -> List[int]:
        """Define get_clean_labels_by_names function."""
        raise NotImplementedError

    @abstractmethod
    def get_labels_by_names(self, names: List[str], **kwargs) -> List[int]:
        """Define get_labels_by_names function."""
        raise NotImplementedError

    @abstractmethod
    def get_image_by_name(self, name: str, **kwargs) -> "tensor":  # type: ignore # noqa
        """Define get_image_by_name function."""
        raise NotImplementedError
